#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
open-MaStR - Main file

Bulk: Download XML-Dump and fill in local SQLite database.
API: Download latest entries using the SOAP-API.

SPDX-License-Identifier: AGPL-3.0-or-later
"""

from open_mastr import Mastr
import os

## specify download parameter

# set custom output path for: csv-export, database, xml-export.
# os.environ['OUTPUT_PATH'] = r"/your/custom/output_path"

# optimize bulk downloads and use the recommended number of processes
# os.environ['USE_RECOMMENDED_NUMBER_OF_PROCESSES'] = "True"

# set up your own number of processes for bulk download
# os.environ['NUMBER_OF_PROCESSES'] = "your_number"

# bulk download
bulk_date = "today"
bulk_cleansing = True
data_bulk = [
    "biomass",
    "combustion",
    "gsgk",
    "hydro",
    "nuclear",
    "solar",
    "storage",
    "wind",
    "balancing_area",
    "electricity_consumer",
    "gas",
    "grid",
    "location",
    "market",
    "permit",
]

# API download
# for parameter explanation see: https://open-mastr.readthedocs.io/en/latest/advanced/#soap-api-download

api_date = "latest"
api_chunksize = 10
api_limit = 10
api_processes = None

data_api = [
    "biomass",
    "combustion",
    "gsgk",
    "hydro",
    "nuclear",
    "solar",
    "storage",
    "wind",
]

api_data_types = ["unit_data", "eeg_data", "kwk_data", "permit_data"]

api_location_types = [
    "location_elec_generation",
    "location_elec_consumption",
    "location_gas_generation",
    "location_gas_consumption",
]

if __name__ == "__main__":
    # instantiate Mastr class
    db = Mastr()

    ## download Markstammdatenregister
    # bulk download
    db.download(method="bulk", data=data_bulk, date=bulk_date, bulk_cleansing=True)

    # API download
    db.download(
        method="API",
        data=data_api,
        date=api_date,
        api_processes=api_processes,
        api_limit=api_limit,
        api_chunksize=api_chunksize,
        api_data_types=api_data_types,
        api_location_types=api_location_types,
    )

    ## export to csv
    """
    Technology-related tables are exported as joined, whereas additional tables
    are duplicated as they are in the database. 
    """
    db.to_csv()



import pytest
import os
from os.path import expanduser
import sys
import random
from os.path import join
from datetime import datetime
import pandas as pd
from open_mastr import Mastr

from open_mastr.utils import orm
from open_mastr.utils.constants import (
    API_LOCATION_TYPES,
    TECHNOLOGIES,
    ADDITIONAL_TABLES,
)
from open_mastr.utils.config import get_data_version_dir, create_data_dir
from open_mastr.utils.helpers import (
    validate_parameter_format_for_download_method,
    validate_parameter_format_for_mastr_init,
    validate_api_credentials,
    transform_data_parameter,
    data_to_include_tables,
    session_scope,
    create_db_query,
    db_query_to_csv,
    reverse_unit_type_map,
)


# Check if db is empty
_db_exists = False
_db_folder_path = os.path.join(
    expanduser("~"), ".open-MaStR", "data", "sqlite"
)  # FIXME: use path in tmpdir when implemented
if os.path.isdir(_db_folder_path):
    for entry in os.scandir(path=_db_folder_path):
        _db_path = os.path.join(_db_folder_path, "open-mastr.db")
        if os.path.getsize(_db_path) > 1000000:  # empty db = 327.7kB < 1 MB
            _db_exists = True


@pytest.fixture
def db():
    return Mastr()


@pytest.fixture
def parameter_dict_working_list():
    parameter_dict_bulk = {
        "method": ["bulk"],
        "data": [
            "wind",
            "solar",
            "biomass",
            "hydro",
            "gsgk",
            "combustion",
            "nuclear",
            "gas",
            "storage",
            "electricity_consumer",
            "location",
            "market",
            "grid",
            "balancing_area",
            "permit",
            "deleted_units",
            "deleted_market_actors",
            "retrofit_units",
            None,
            ["wind", "solar"],
        ],
        "date": ["today", "20200108", "existing"],
        "bulk_cleansing": [True, False],
        "api_processes": [None],
        "api_limit": [50],
        "api_chunksize": [1000],
        "api_data_types": [None],
        "api_location_types": [None],
    }

    parameter_dict_API = {
        "method": ["API"],
        "data": [
            "wind",
            "solar",
            "biomass",
            "hydro",
            "gsgk",
            "combustion",
            "nuclear",
            "storage",
            "location",
            "permit",
            None,
            ["wind", "solar"],
        ],
        "date": [None, datetime(2022, 2, 2), "latest"],
        "bulk_cleansing": [True],
        "api_processes": [None]
        if sys.platform not in ["linux2", "linux"]
        else [2, 20, None, "max"],
        "api_limit": [15, None],
        "api_chunksize": [20],
        "api_data_types": [
            ["unit_data", "eeg_data", "kwk_data", "permit_data"],
            ["unit_data"],
            None,
        ],
        "api_location_types": [
            ["location_elec_generation", "location_elec_consumption"],
            [
                "location_elec_generation",
                "location_elec_consumption",
                "location_gas_generation",
                "location_gas_consumption",
            ],
            None,
        ],
    }
    return [parameter_dict_bulk, parameter_dict_API]


@pytest.fixture
def parameter_dict_not_working():
    parameter_dict = {
        "method": [5, "BULK", "api"],
        "data": ["wint", "Solar", "biomasse", 5, []],
        "date": [124, "heute", 123],
        "bulk_cleansing": ["cleansing", 4, None],
        "api_processes": ["20", "None"],
        "api_limit": ["15", "None"],
        "api_chunksize": ["20"],
        "api_data_types": ["unite_data", 5, []],
        "api_location_types": ["locatione_elec_generation", 5, []],
    }
    return parameter_dict


def test_Mastr_validate_working_parameter(parameter_dict_working_list):
    for parameter_dict_working in parameter_dict_working_list:
        parameter_dict = {
            key: parameter_dict_working[key][0] for key in parameter_dict_working
        }

        # working parameters
        for key in list(parameter_dict_working.keys()):
            for value in parameter_dict_working[key]:
                parameter_dict[key] = value
                (
                    method,
                    data,
                    date,
                    bulk_cleansing,
                    api_processes,
                    api_limit,
                    api_chunksize,
                    api_data_types,
                    api_location_types,
                ) = get_parameters_from_parameter_dict(parameter_dict)

                assert (
                    validate_parameter_format_for_download_method(
                        method,
                        data,
                        date,
                        bulk_cleansing,
                        api_processes,
                        api_limit,
                        api_chunksize,
                        api_data_types,
                        api_location_types,
                    )
                    is None
                )


def test_Mastr_validate_not_working_parameter(
    parameter_dict_working_list, parameter_dict_not_working
):
    for parameter_dict_working in parameter_dict_working_list:
        parameter_dict_initial = {
            key: parameter_dict_working[key][0] for key in parameter_dict_working
        }

        # not working parameters
        for key in list(parameter_dict_not_working.keys()):
            for value in parameter_dict_not_working[key]:
                # reset parameter_dict so that all parameters are working except one
                parameter_dict = parameter_dict_initial.copy()
                parameter_dict[key] = value
                (
                    method,
                    data,
                    date,
                    bulk_cleansing,
                    api_processes,
                    api_limit,
                    api_chunksize,
                    api_data_types,
                    api_location_types,
                ) = get_parameters_from_parameter_dict(parameter_dict)
                with pytest.raises(ValueError):
                    validate_parameter_format_for_download_method(
                        method,
                        data,
                        date,
                        bulk_cleansing,
                        api_processes,
                        api_limit,
                        api_chunksize,
                        api_data_types,
                        api_location_types,
                    )


def get_parameters_from_parameter_dict(parameter_dict):
    method = parameter_dict["method"]
    data = parameter_dict["data"]
    date = parameter_dict["date"]
    bulk_cleansing = parameter_dict["bulk_cleansing"]
    api_processes = parameter_dict["api_processes"]
    api_limit = parameter_dict["api_limit"]
    api_chunksize = parameter_dict["api_chunksize"]
    api_data_types = parameter_dict["api_data_types"]
    api_location_types = parameter_dict["api_location_types"]
    return (
        method,
        data,
        date,
        bulk_cleansing,
        api_processes,
        api_limit,
        api_chunksize,
        api_data_types,
        api_location_types,
    )


def test_validate_parameter_format_for_mastr_init(db):
    engine_list_working = ["sqlite", db.engine]
    engine_list_failing = ["HI", 12]

    for engine in engine_list_working:
        assert validate_parameter_format_for_mastr_init(engine) is None

    for engine in engine_list_failing:
        with pytest.raises(ValueError):
            validate_parameter_format_for_mastr_init(engine)


def test_transform_data_parameter():
    (data, api_data_types, api_location_types, harm_log,) = transform_data_parameter(
        method="API",
        data=["wind", "location"],
        api_data_types=["eeg_data"],
        api_location_types=None,
    )

    assert data == ["wind"]
    assert api_data_types == ["eeg_data"]
    assert api_location_types == API_LOCATION_TYPES
    assert harm_log == ["location"]


def test_validate_api_credentials():
    validate_api_credentials()


def test_data_to_include_tables():
    # Prepare
    include_tables_list = [
        "anlageneegwind",
        "einheitenwind",
        "anlageneegwasser",
        "einheitenwasser",
    ]
    include_tables_str = ["einheitenstromverbraucher"]

    map_to_db_table_list = ["market_actors", "market_roles"]
    map_to_db_table_str = ["locations_extended"]

    # Assert
    assert include_tables_list == data_to_include_tables(
        data=["wind", "hydro"], mapping="write_xml"
    )
    assert include_tables_str == data_to_include_tables(
        data=["electricity_consumer"], mapping="write_xml"
    )
    assert map_to_db_table_list == data_to_include_tables(
        data=["market"], mapping="export_db_tables"
    )
    assert map_to_db_table_str == data_to_include_tables(
        data=["location"], mapping="export_db_tables"
    )


def test_data_to_include_tables_error():
    # test for non-existent 'mapping' parameter input
    with pytest.raises(
        NotImplementedError,
        match="This function is only implemented for 'write_xml' and 'export_db_tables',"
        " please specify when calling the function.",
    ):
        data_to_include_tables(data=["wind", "hydro"], mapping="X32J_22")


@pytest.mark.skipif(
    not _db_exists,
    reason="The database is smaller than 1 MB, thus suspected to be empty or non-existent.",
)
def test_db_query_to_csv(tmpdir, engine):
    """
    The test checks for 2 random tech and 2 random additional tables:
    1. If csv's have been created and encoded in 'utf-8' and are not empty
    2. For techs, if limited (limit=60) EinheitMastrNummer in basic_units are included in CSV file
    3. For additional_tables, if csv is not empty

    Parameters
    ----------
    tmpdir - temporary directory to test export
    engine - sqlite engine from conftest.py

    Returns
    -------

    """
    unit_type_map_reversed = reverse_unit_type_map()

    # FIXME: Define path to tmpdir (pytest internal temporary dir)
    # to not delete actual data export, when test is run locally
    #  Use the parameter that will be implemented in #394
    # create data dir
    create_data_dir()

    techs = random.sample(TECHNOLOGIES, k=2)
    addit_tables = random.sample(ADDITIONAL_TABLES, k=2)

    with session_scope(engine=engine) as session:
        for tech in techs:
            db_query_to_csv(
                db_query=create_db_query(tech=tech, limit=60, engine=engine),
                data_table=tech,
                chunksize=20,
            )

            # Test if LIMITED EinheitMastrNummer in basic_units are included in CSV file
            csv_path = join(
                get_data_version_dir(),
                f"bnetza_mastr_{tech}_raw.csv",
            )
            # check if table has been created and encoding is correct
            df_tech = pd.read_csv(
                csv_path, index_col="EinheitMastrNummer", encoding="utf-8"
            )

            # check whether table is empty (returns True if it is)
            assert False == df_tech.empty

            units = session.query(orm.BasicUnit.EinheitMastrNummer).filter(
                orm.BasicUnit.Einheittyp == unit_type_map_reversed[tech]
            )
            set_MastrNummer = {unit.EinheitMastrNummer for unit in units}
            for idx in df_tech.index:
                assert idx in set_MastrNummer

            # FIXME: delete when tmpdir is implemented
            # remove file in data folder
            os.remove(csv_path)

        for addit_table in addit_tables:
            csv_path = join(
                get_data_version_dir(),
                f"bnetza_mastr_{addit_table}_raw.csv",
            )

            db_query_to_csv(
                db_query=create_db_query(
                    additional_table=addit_table, limit=60, engine=engine
                ),
                data_table=addit_table,
                chunksize=20,
            )

            # check if table has been created and encoding is correct
            df_at = pd.read_csv(csv_path, encoding="utf-8")

            # check if table is empty (returns True if it is)
            assert False == df_at.empty

            # FIXME: delete when tmpdir is implemented
            # remove file in data folder
            os.remove(csv_path)

    # FIXME: delete when tmpdir is implemented
    # delete empty data dir
    os.rmdir(get_data_version_dir())


def test_save_metadata():
    # FIXME: implement in #386
    pass



from open_mastr.soap_api.download import MaStRAPI, MaStRDownload, flatten_dict
import pytest
import datetime


@pytest.fixture
def mastr_api_fake_credentials():
    return MaStRAPI(user="testuser", key="testpassword")


@pytest.fixture
def mastr_api():
    return MaStRAPI()


@pytest.fixture
def mastr_download():
    return MaStRDownload()


@pytest.mark.dependency(name="db_reachable")
def test_soap_wrapper_connection(mastr_api_fake_credentials):
    mastr_api = mastr_api_fake_credentials
    lokale_uhrzeit = mastr_api.GetLokaleUhrzeit()

    assert lokale_uhrzeit["Ergebniscode"] == "OK"


@pytest.mark.dependency(depends=["db_reachable"])
def test_soap_wrapper_contingent(mastr_api):
    contingent = mastr_api.GetAktuellerStandTageskontingent()

    assert contingent["Ergebniscode"] == "OK"


@pytest.mark.dependency(depends=["db_reachable"])
def test_soap_wrapper_power_plant_list(mastr_api):
    response = mastr_api.GetGefilterteListeStromErzeuger(limit=1)

    for key in [
        "EinheitMastrNummer",
        "DatumLetzeAktualisierung",
        "Name",
        "Einheitart",
        "Einheittyp",
        "Standort",
        "Bruttoleistung",
        "Erzeugungsleistung",
        "EinheitBetriebsstatus",
        "Anlagenbetreiber",
        "EegMastrNummer",
        "KwkMastrNummer",
        "SpeMastrNummer",
        "GenMastrNummer",
        "BestandsanlageMastrNummer",
        "NichtVorhandenInMigriertenEinheiten",
    ]:
        for einheit in response["Einheiten"]:
            assert key in einheit

    assert response["Ergebniscode"] == "OkWeitereDatenVorhanden"


def test_basic_unit_data(mastr_download):
    data = [
        unit
        for sublist in mastr_download.basic_unit_data(data="nuclear", limit=1)
        for unit in sublist
    ]

    assert len(data) == 1
    assert data[0]["Einheittyp"] == "Kernenergie"


def test_additional_data_nuclear(mastr_download):

    data_fcns = [
        ("SME963513379837", "extended_unit_data"),
        ("SGE951929415553", "permit_unit_data"),
    ]

    for mastr_nummer, data_fcn in data_fcns:
        units_downloaded, units_missed = mastr_download.additional_data(
            "nuclear", [mastr_nummer], data_fcn
        )

        assert len(units_downloaded) + len(units_missed) == 1


def test_additional_data_biomass(mastr_download):

    data_fcns = [
        ("SEE936595511945", "extended_unit_data"),
        ("EEG929630520224", "extended_unit_data"),
        ("KWK998480117397", "kwk_unit_data"),
        ("SGE974254715891", "permit_unit_data"),
    ]

    for mastr_nummer, data_fcn in data_fcns:
        units_downloaded, units_missed = mastr_download.additional_data(
            "biomass", [mastr_nummer], data_fcn
        )

        assert len(units_downloaded) + len(units_missed) == 1


def test_flatten_dict():
    data_before_flatten = [
        {
            "Frist": {"Wert": datetime.date(2017, 6, 15), "NichtVorhanden": False},
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": {"Wert": None, "NichtVorhanden": False},
            "Registrierungsdatum": datetime.date(2019, 9, 11),
            "VerknuepfteEinheiten": [
                {
                    "MaStRNummer": "SEE900290686291",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                }
            ],
        },
        {
            "Frist": {"Wert": datetime.date(2018, 3, 27), "NichtVorhanden": False},
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": {"Wert": None, "NichtVorhanden": False},
            "Registrierungsdatum": datetime.date(2019, 2, 4),
            "VerknuepfteEinheiten": [
                {
                    "MaStRNummer": "SEE908999761141",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
                {
                    "MaStRNummer": "SEE978389475514",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
                {
                    "MaStRNummer": "SEE960287756734",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
            ],
        },
        {
            "Frist": {"Wert": datetime.date(2018, 3, 23), "NichtVorhanden": False},
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": {"Wert": None, "NichtVorhanden": False},
            "Registrierungsdatum": datetime.date(2019, 2, 15),
            "VerknuepfteEinheiten": [
                {
                    "MaStRNummer": "SEE902566900605",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
                {
                    "MaStRNummer": "SEE945851284479",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
                {
                    "MaStRNummer": "SEE975973981666",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
            ],
        },
        {
            "Frist": {"Wert": datetime.date(2016, 1, 23), "NichtVorhanden": False},
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": {"Wert": None, "NichtVorhanden": False},
            "Registrierungsdatum": datetime.date(2019, 2, 7),
            "VerknuepfteEinheiten": [
                {
                    "MaStRNummer": "SEE969499157391",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
                {
                    "MaStRNummer": "SEE949984955732",
                    "Einheittyp": "Windeinheit",
                    "Einheitart": "Stromerzeugungseinheit",
                },
            ],
        },
    ]
    data_after_flatten = [
        {
            "Frist": datetime.date(2017, 6, 15),
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": None,
            "Registrierungsdatum": datetime.date(2019, 9, 11),
            "VerknuepfteEinheiten": "SEE900290686291",
        },
        {
            "Frist": datetime.date(2018, 3, 27),
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": None,
            "Registrierungsdatum": datetime.date(2019, 2, 4),
            "VerknuepfteEinheiten": "SEE908999761141, SEE978389475514, SEE960287756734",
        },
        {
            "Frist": datetime.date(2018, 3, 23),
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": None,
            "Registrierungsdatum": datetime.date(2019, 2, 15),
            "VerknuepfteEinheiten": "SEE902566900605, SEE945851284479, SEE975973981666",
        },
        {
            "Frist": datetime.date(2016, 1, 23),
            "WasserrechtsNummer": None,
            "WasserrechtAblaufdatum": None,
            "Registrierungsdatum": datetime.date(2019, 2, 7),
            "VerknuepfteEinheiten": "SEE969499157391, SEE949984955732",
        },
    ]
    assert flatten_dict(data_before_flatten) == data_after_flatten



import os
import shutil
import time
from importlib.metadata import PackageNotFoundError, version
from zipfile import BadZipfile, ZipFile

import numpy as np
import requests
from tqdm import tqdm

# setup logger
from open_mastr.utils.config import setup_logger

try:
    USER_AGENT = (
        f"open-mastr/{version('open-mastr')} python-requests/{version('requests')}"
    )
except PackageNotFoundError:
    USER_AGENT = "open-mastr"
log = setup_logger()


def gen_version(
    when: time.struct_time = time.localtime(), use_version: str = "current"
) -> str:
    """
    Generates the current version.

    The version number is determined according to a fixed release cycle,
    which is by convention in sync with the changes to other german regulatory
    frameworks of the energy such as GeLI Gas and GPKE.

    The release schedule is twice per year on 1st of April and October.
    The version number is determined by the year of release and the running
    number of the release, i.e. the release on April 1st is release 1,
    while the release in October is release 2.

    Further, the release happens during the day, so on the day of the
    changeover, the exported data will still be in the old version/format.

    see <https://www.marktstammdatenregister.de/MaStRHilfe/files/webdienst/Release-Termine.pdf>

    Examples:
    2024-01-01 = version 23.2
    2024-04-01 = version 23.2
    2024-04-02 = version 24.1
    2024-09-30 = version 24.1
    2024-10-01 = version 24.1
    2024-10-02 = version 24.2
    2024-31-12 = version 24.2
    """

    year = when.tm_year
    release = 1

    if when.tm_mon < 4 or (when.tm_mon == 4 and when.tm_mday == 1):
        year = year - 1
        release = 2
    elif when.tm_mon > 10 or (when.tm_mon == 10 and when.tm_mday > 1):
        release = 2

    # Change to MaStR version number that was used before
    # For example: 24.1 -> 23.2
    if use_version == "before":
        if release == 1:
            year = year - 1
            release = 2
        else:
            release = 1
    # Change to MaStR version number that was used afterwards
    # For example: 24.1 -> 24.2
    elif use_version == "after":
        if release == 2:
            year = year + 1
            release = 1
        else:
            release = 2

    # only the last two digits of the year are used
    year = str(year)[-2:]
    return f"{year}.{release}"


def gen_url(when: time.struct_time = time.localtime(), use_version="current") -> str:
    """Generates the download URL for the specified date.

    Note that not all dates are archived on the website.
    Normally only today is available, the export is usually made
    between 02:00 and 04:00, which means before 04:00 the current data may not
    yet be available and the download could fail.

    Note also that this function will not be able to generate URLs for dates
    before 2024 because a different URL scheme was used then which had some random
    data embedded in the name to make it harder to automate downloads.


    Args:
        when (time.struct_time, optional): Time object used to generate url. Defaults to time.localtime().
        use_version (str, optional): One of "current", "before", "after". "current" will generate the url
        for the expected MaStR version. "before" will generate the url for the previous MaStR version.
        "after" will generate the url for the subsequent MaStR version.

        "current": Gesamtdatenexport_20250403_25.1.zip
        "before": Gesamtdatenexport_20250403_24.2.zip
        "after": Gesamtdatenexport_20250403_25.2.zip

        Defaults to "current".
    """
    version = gen_version(when, use_version)
    date = time.strftime("%Y%m%d", when)

    return f"https://download.marktstammdatenregister.de/Gesamtdatenexport_{date}_{version}.zip"


def download_xml_Mastr(
    save_path: str, bulk_date_string: str, xml_folder_path: str
) -> None:
    """Downloads the zipped MaStR.

    Parameters
    -----------
    save_path: str
        The path where the downloaded MaStR zipped folder will be saved.
    """

    if os.path.exists(save_path):
        try:
            _ = ZipFile(save_path)
        except BadZipfile:
            log.info(f"Bad Zip file is deleted: {save_path}")
            os.remove(save_path)
        else:
            print("MaStR already downloaded.")
            return None

    if bulk_date_string != "today":
        raise OSError(
            "There exists no file for given date. MaStR can only be downloaded "
            "from the website if today's date is given."
        )
    shutil.rmtree(xml_folder_path, ignore_errors=True)
    os.makedirs(xml_folder_path, exist_ok=True)

    print_message = (
        "Download has started, this can take several minutes."
        "The download bar is only a rough estimate."
    )
    warning_message = (
        "Warning: The servers from MaStR restrict the download speed."
        " You may want to download it another time."
    )
    print(print_message)

    now = time.localtime()
    url = gen_url(now)

    time_a = time.perf_counter()
    r = requests.get(url, stream=True, headers={"User-Agent": USER_AGENT})
    if r.status_code == 404:
        log.warning(
            "Download file was not found. Assuming that the new file was not published yet and retrying with yesterday."
        )
        now = time.localtime(
            time.mktime(now) - (24 * 60 * 60)
        )  # subtract 1 day from the date
        url = gen_url(now)
        r = requests.get(url, stream=True, headers={"User-Agent": USER_AGENT})
    if r.status_code == 404:
        url = gen_url(now, use_version="before")  # Use lower MaStR Version
        log.warning(
            f"Download file was not found. Assuming that the version of MaStR has changed and retrying with download link: {url}"
        )
        r = requests.get(url, stream=True, headers={"User-Agent": USER_AGENT})
    if r.status_code == 404:
        url = gen_url(now, use_version="after")  # Use higher MaStR Version
        log.warning(
            f"Download file was not found. Assuming that the version of MaStR has changed and retrying with download link: {url}"
        )
        r = requests.get(url, stream=True, headers={"User-Agent": USER_AGENT})

    if r.status_code == 404:
        log.error("Could not download file: download URL not found")
        return

    total_length = int(18000 * 1024 * 1024)
    with (
        open(save_path, "wb") as zfile,
        tqdm(desc=save_path, total=(total_length / 1024 / 1024), unit="") as bar,
    ):
        for chunk in r.iter_content(chunk_size=1024 * 1024):
            # chunk size of 1024 * 1024 needs 9min 11 sek = 551sek
            # chunk size of 1024 needs 9min 11 sek as well
            if chunk:
                zfile.write(chunk)
                zfile.flush()
            bar.update()
            # if the rate falls below 100 kB/s -> prompt warning
            if bar.format_dict["rate"] and bar.format_dict["rate"] < 2:
                bar.set_postfix_str(s=warning_message)
            else:
                # remove warning
                bar.set_postfix_str(s="")
    time_b = time.perf_counter()
    print(f"Download is finished. It took {int(np.around(time_b - time_a))} seconds.")
    print(f"MaStR was successfully downloaded to {xml_folder_path}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Matches wind turbines to the wind_turbine_library
https://openenergy-platform.org/dataedit/view/supply/wind_turbine_library
"""

__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "solar-c"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
__version__ = "v0.10.0"


import pandas as pd
import os


def read_csv_turbine(csv_name):
    turbines = pd.read_csv(
        csv_name,
        header=0,
        encoding="utf-8",
        sep=",",
        error_bad_lines=True,
        index_col=False,
        dtype={
            "index": int,
            "id": int,
            "turbine_id": int,
            "manufacturer": str,
            "name": str,
            "turbine_type": str,
            "nominal_power": str,
            "rotor_diamter": str,
            "rotor_area": str,
            "hub_height": str,
            "max_speed_drive": str,
            "wind_class_iec": str,
            "wind_zone_dibt": str,
            "power_density": str,
            "power_density_2": str,
            "calculated": str,
            "has_power_curve": str,
            "power_curve_wind_speeds": str,
            "power_curve_values": str,
            "has_cp_curve": str,
            "power_coefficient_curve_wind_speeds": str,
            "power_coefficient_curve_values": str,
            "has_ct_curve": str,
            "thrust_coefficient_curve_wind_speeds": str,
            "thrust_coefficient_curve_values": str,
            "source": str,
        },
    )
    return turbines


def create_dataset(df):
    types = []
    for i, r in df.iterrows():
        types.append(prepare_turbine_type(r))
    df.insert(6, "turbine_type_v2", types)
    write_to_csv(df, "turbine_library_t.csv")


def write_to_csv(df, path):
    with open(path, mode="a", encoding="utf-8") as file:
        df.to_csv(
            file,
            sep=",",
            mode="a",
            header=file.tell() == 0,
            line_terminator="\n",
            encoding="utf-8",
        )


def prepare_turbine_type(turbine):
    nom_pow = turbine.nominal_power
    diam = turbine.rotor_diameter
    man = get_manufacturer_short(turbine.manufacturer, nom_pow, diam)
    type_name = man + "-" + str(diam) + "_" + str(int(nom_pow))
    return type_name


def get_manufacturer_short(manufacturer, nom_pow, diam):
    man = ""
    if manufacturer == "Nordex":
        man = "N"
        if int(nom_pow) == 3000 or int(nom_pow) == 1500:
            if (
                int(diam) == 140
                or int(diam) == 132
                or int(diam) == 125
                or int(diam) == 116
                or int(diam) == 100
                or int(diam) == 82
                or int(diam) == 77
                or int(diam) == 70
            ):
                man = "AW"
    elif manufacturer == "Adwen/Areva":
        man = "AD"
    elif manufacturer == "Senvion/REpower":
        man = "S"
        if int(nom_pow) == 2050 or int(nom_pow) == 2000:
            man = "MM"
    elif manufacturer == "Enercon":
        man = "E"
    elif manufacturer == "Siemens":
        man = "SWT"
    elif manufacturer == "Vestas":
        man = "V"
    elif manufacturer == "Vensys":
        man = "VS"
    elif manufacturer == "GE Wind":
        man = "GE"
    elif manufacturer == "Eno":
        man = "ENO"
    elif manufacturer == "aerodyn":
        man = "SCD"
    return man



"""
The conftest.py file serves as a means of providing fixtures for an entire directory.
Fixtures defined in a conftest.py can be used by any test in that package without
needing to import them (pytest will automatically discover them).

You can have multiple nested directories/packages containing your tests,
and each directory can have its own conftest.py with its own fixtures,
adding on to the ones provided by the conftest.py files in parent directories.

https://docs.pytest.org/en/7.2.x/reference/fixtures.html
"""

import pytest
from open_mastr import Mastr

from open_mastr.utils.config import get_project_home_dir
from open_mastr.utils.helpers import create_database_engine
import os


@pytest.fixture(scope="function")
def make_Mastr_class():
    """
    Factory to create different Mastr class objects.

    Parameters
    ----------
    engine_type: str
        Define type of engine, for details see
        :meth: `~.open_mastr.utils.helpers.create_database_engine`

    Returns
    -------
        Mastr class object
    """

    def _make_Mastr_class(engine_type):
        return Mastr(engine=engine_type)

    return _make_Mastr_class


@pytest.fixture
def engine():
    return create_database_engine(
        "sqlite", os.path.join(get_project_home_dir(), "data", "sqlite")
    )



import os
import json
import sys
from contextlib import contextmanager
from datetime import date, datetime
from warnings import warn

import dateutil
import sqlalchemy
from sqlalchemy.sql import insert, literal_column, text
from dateutil.parser import parse
from sqlalchemy import create_engine
from sqlalchemy.orm import Query, sessionmaker

import pandas as pd
from tqdm import tqdm
from open_mastr.soap_api.metadata.create import create_datapackage_meta_json
from open_mastr.utils import orm
from open_mastr.utils.config import (
    get_filenames,
    get_data_version_dir,
    column_renaming,
)

from open_mastr.soap_api.download import MaStRAPI, log
from open_mastr.utils.constants import (
    BULK_DATA,
    TECHNOLOGIES,
    API_DATA,
    API_DATA_TYPES,
    API_LOCATION_TYPES,
    BULK_INCLUDE_TABLES_MAP,
    BULK_ADDITIONAL_TABLES_CSV_EXPORT_MAP,
    ORM_MAP,
    UNIT_TYPE_MAP,
    ADDITIONAL_TABLES,
    TRANSLATIONS,
)


def chunks(lst, n):
    """Yield successive n-sized chunks from lst.

    `Credits
    <https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks>`_
    """
    length = lst.count() if isinstance(lst, Query) else len(lst)
    for i in range(0, length, n):
        yield lst[i : i + n]


def create_database_engine(engine, sqlite_db_path) -> sqlalchemy.engine.Engine:
    if engine == "sqlite":
        sqlite_database_path = os.environ.get(
            "SQLITE_DATABASE_PATH",
            os.path.join(sqlite_db_path, "open-mastr.db"),
        )
        db_url = f"sqlite:///{sqlite_database_path}"
        return create_engine(db_url)

    if type(engine) == sqlalchemy.engine.Engine:
        return engine


def parse_date_string(bulk_date_string: str) -> str:
    if bulk_date_string == "today":
        return date.today().strftime("%Y%m%d")
    else:
        return parse(bulk_date_string).strftime("%Y%m%d")


def validate_parameter_format_for_mastr_init(engine) -> None:
    if engine not in ["sqlite"] and not isinstance(engine, sqlalchemy.engine.Engine):
        raise ValueError(
            "parameter engine has to be either 'sqlite' "
            "or an sqlalchemy.engine.Engine object."
        )


def validate_parameter_format_for_download_method(
    method,
    data,
    date,
    bulk_cleansing,
    api_processes,
    api_limit,
    api_chunksize,
    api_data_types,
    api_location_types,
    **kwargs,
) -> None:
    if "technology" in kwargs:
        data = kwargs["technology"]
        warn("'technology' parameter is deprecated. Use 'data' instead")
    if "bulk_date" in kwargs:
        date = kwargs["bulk_date"]
        warn("'bulk_date' parameter is deprecated. Use 'date' instead")
    if "api_date" in kwargs:
        date = kwargs["api_date"]
        warn("'api_date' parameter is deprecated. Use 'date' instead")

    validate_parameter_method(method)
    validate_parameter_data(method, data)
    validate_parameter_date(method, date)
    validate_parameter_bulk_cleansing(bulk_cleansing)
    validate_parameter_api_processes(api_processes)
    validate_parameter_api_limit(api_limit)
    validate_parameter_api_chunksize(api_chunksize)
    validate_parameter_api_data_types(api_data_types)
    validate_parameter_api_location_types(api_location_types)

    raise_warning_for_invalid_parameter_combinations(
        method,
        bulk_cleansing,
        date,
        api_processes,
        api_data_types,
        api_location_types,
        api_limit,
        api_chunksize,
    )


def validate_parameter_method(method) -> None:
    if method not in ["bulk", "API"]:
        raise ValueError("parameter method has to be either 'bulk' or 'API'.")


def validate_parameter_api_location_types(api_location_types) -> None:
    if not isinstance(api_location_types, list) and api_location_types is not None:
        raise ValueError("parameter api_location_types has to be a list or 'None'.")

    if isinstance(api_location_types, list):
        if not api_location_types:  # api_location_types == []
            raise ValueError("parameter api_location_types cannot be an empty list!")
        for value in api_location_types:
            if value not in API_LOCATION_TYPES:
                raise ValueError(
                    f"list entries of api_data_types have to be in {API_LOCATION_TYPES}."
                )


def validate_parameter_api_data_types(api_data_types) -> None:
    if not isinstance(api_data_types, list) and api_data_types is not None:
        raise ValueError("parameter api_data_types has to be a list or 'None'.")

    if isinstance(api_data_types, list):
        if not api_data_types:  # api_data_types == []
            raise ValueError("parameter api_data_types cannot be an empty list!")
        for value in api_data_types:
            if value not in API_DATA_TYPES:
                raise ValueError(
                    f"list entries of api_data_types have to be in {API_DATA_TYPES}."
                )


def validate_parameter_api_chunksize(api_chunksize) -> None:
    if not isinstance(api_chunksize, int) and api_chunksize is not None:
        raise ValueError("parameter api_chunksize has to be an integer or 'None'.")


def validate_parameter_bulk_cleansing(bulk_cleansing) -> None:
    if type(bulk_cleansing) != bool:
        raise ValueError("parameter bulk_cleansing has to be boolean")


def validate_parameter_api_limit(api_limit) -> None:
    if not isinstance(api_limit, int) and api_limit is not None:
        raise ValueError("parameter api_limit has to be an integer or 'None'.")


def validate_parameter_date(method, date) -> None:
    if date is None:  # default
        return
    if method == "bulk":
        if date not in ["today", "existing"]:
            try:
                _ = parse(date)
            except (dateutil.parser._parser.ParserError, TypeError) as e:
                raise ValueError(
                    "Parameter date has to be a proper date in the format yyyymmdd"
                    "or 'today' for bulk method."
                ) from e
    elif method == "API":
        if not isinstance(date, datetime) and date != "latest":
            raise ValueError(
                "parameter api_date has to be 'latest' or a datetime object or 'None'"
                " for API method."
            )


def validate_parameter_api_processes(api_processes) -> None:
    if api_processes not in ["max", None] and not isinstance(api_processes, int):
        raise ValueError(
            "parameter api_processes has to be 'max' or an integer or 'None'"
        )
    if api_processes == "max" or isinstance(api_processes, int):
        system = sys.platform
        if system not in ["linux2", "linux"]:
            raise ValueError(
                "The functionality of multiprocessing only works on Linux based systems. "
                "On your system, the parameter api_processes has to be 'None'."
            )


def validate_parameter_data(method, data) -> None:
    if not isinstance(data, (str, list)) and data is not None:
        raise ValueError("parameter data has to be a string, list, or None")
    if isinstance(data, str):
        data = [data]
    if isinstance(data, list):
        if not data:  # data == []
            raise ValueError("parameter data cannot be an empty list!")
        for value in data:
            if method == "bulk" and value not in BULK_DATA:
                raise ValueError(
                    f"Allowed values for parameter data with bulk method are {BULK_DATA}"
                )
            if method == "API" and value not in API_DATA:
                raise ValueError(
                    f"Allowed values for parameter data with API method are {API_DATA}"
                )
            if method == "csv_export" and value not in TECHNOLOGIES + ADDITIONAL_TABLES:
                raise ValueError(
                    "Allowed values for CSV export are "
                    f"{TECHNOLOGIES} or {ADDITIONAL_TABLES}"
                )


def raise_warning_for_invalid_parameter_combinations(
    method,
    bulk_cleansing,
    date,
    api_processes,
    api_data_types,
    api_location_types,
    api_limit,
    api_chunksize,
):
    if method == "API" and bulk_cleansing is not True:
        warn(
            "For method = 'API', bulk download related parameters "
            "(with prefix bulk_) are ignored."
        )

    if method == "bulk" and (
        (
            any(
                parameter is not None
                for parameter in [
                    api_processes,
                    api_data_types,
                    api_location_types,
                ]
            )
            or api_limit != 50
            or api_chunksize != 1000
        )
    ):
        warn(
            "For method = 'bulk', API related parameters (with prefix api_) are ignored."
        )


def transform_data_parameter(
    method, data, api_data_types, api_location_types, **kwargs
):
    """
    Parse input parameters related to data as lists. Harmonize variables for later use.
    Data output depends on the possible data types of chosen method.
    """
    data = kwargs.get("technology", data)
    # parse parameters as list
    if isinstance(data, str):
        data = [data]
    elif data is None:
        data = BULK_DATA if method == "bulk" else API_DATA
    if api_data_types is None:
        api_data_types = API_DATA_TYPES
    if api_location_types is None:
        api_location_types = API_LOCATION_TYPES

    # data input harmonisation
    harmonisation_log = []
    if method == "API" and "permit" in data:
        data.remove("permit")
        api_data_types.append(
            "permit_data"
        ) if "permit_data" not in api_data_types else api_data_types
        harmonisation_log.append("permit")

    if method == "API" and "location" in data:
        data.remove("location")
        api_location_types = API_LOCATION_TYPES
        harmonisation_log.append("location")

    return data, api_data_types, api_location_types, harmonisation_log


def transform_date_parameter(self, method, date, **kwargs):
    if method == "bulk":
        date = kwargs.get("bulk_date", date)
        date = "today" if date is None else date
        if date == "existing":
            existing_files_list = os.listdir(
                os.path.join(self.output_dir, "data", "xml_download")
            )
            if not existing_files_list:
                date = "today"
                print(
                    "By choosing `date`='existing' you want to use an existing "
                    "xml download."
                    "However no xml_files were downloaded yet. The parameter `date` is"
                    "therefore set to 'today'."
                )
            # we assume that there is only one file in the folder which is the
            # zipped xml folder
            date = existing_files_list[0].split("_")[1].split(".")[0]
    elif method == "API":
        date = kwargs.get("api_date", date)

    return date


@contextmanager
def session_scope(engine):
    """Provide a transactional scope around a series of operations."""
    Session = sessionmaker(bind=engine)
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


def print_api_settings(
    harmonisation_log,
    data,
    date,
    api_data_types,
    api_chunksize,
    api_limit,
    api_processes,
    api_location_types,
):
    print(
        f"Downloading with soap_API.\n\n   -- API settings --  \nunits after date: "
        f"{date}\nunit download limit per data: "
        f"{api_limit}\nparallel_processes: {api_processes}\nchunksize: "
        f"{api_chunksize}\ndata_api: {data}"
    )
    if "permit" in harmonisation_log:
        print(
            f"data_types: {api_data_types}" "\033[31m",
            "Attention, 'permit_data' was automatically set in api_data_types, "
            "as you defined 'permit' in parameter data_api.",
            "\033[m",
        )

    else:
        print(f"data_types: {api_data_types}")

    if "location" in harmonisation_log:
        print(
            "location_types:",
            "\033[31m",
            "Attention, 'location' is in parameter data. location_types are set to",
            "\033[m",
            f"{api_location_types}"
            "\n                 If you want to change location_types, please remove 'location' "
            "from data_api and specify api_location_types."
            "\n   ------------------  \n",
        )

    else:
        print(
            f"location_types: {api_location_types}",
            "\n   ------------------  \n",
        )


def validate_api_credentials() -> None:
    mastr_api = MaStRAPI()
    assert mastr_api.GetAktuellerStandTageskontingent()["Ergebniscode"] == "OK"


def data_to_include_tables(data: list, mapping: str = None) -> list:
    """
    Convert user input 'data' to the list 'include_tables'.
    It contains file names from zipped bulk download, if mapping="write_xml".
    It contains database table names, if mapping="export_db_tables".
    Parameters
    ----------
    data: list
        The user input for data selection
    mapping: str
        Specify the mapping dict for the function and thus the list output.
    Returns
    -------
    list
        List of file names | List of database table names
    -------

    """
    if mapping == "write_xml":
        # Map data selection to include tables in xml
        include_tables = [
            table for tech in data for table in BULK_INCLUDE_TABLES_MAP[tech]
        ]
        return include_tables

    if mapping == "export_db_tables":
        # Map data selection to include tables for csv export
        include_tables = [
            table
            for possible_data_bulk in data
            for table in BULK_ADDITIONAL_TABLES_CSV_EXPORT_MAP[possible_data_bulk]
        ]
        return include_tables

    raise NotImplementedError(
        "This function is only implemented for 'write_xml' and 'export_db_tables', "
        "please specify when calling the function."
    )


def reverse_unit_type_map():
    return {v: k for k, v in UNIT_TYPE_MAP.items()}


# EXPORT RELEVANT FUNCTIONS


def create_db_query(
    tech=None,
    additional_table=None,
    additional_data=["unit_data", "eeg_data", "kwk_data", "permit_data"],
    limit=None,
    engine=None,
):
    """
    Create a database query to export a snapshot MaStR data from database to CSV.

    For technologies, during the query creation, additional available data is joined on
    list of basic units. A query for a single technology is created separately because
    of multiple non-overlapping columns. Duplicate columns for a single technology
    (a results on data from different sources) are suffixed.

    The data in the database probably has duplicates because
    of the history how data was collected in the
    Marktstammdatenregister.

    Along with the data, metadata is saved in the file `datapackage.json`.

    Parameters
    ----------
    technology: `list` of `str`
        See list of available technologies in
        `open_mastr.utils.constants.TECHNOLOGIES`
    additional_table: `list` of `str`
        See list of available technologies or additional tables in
        `open_mastr.utils.constants.ADDITIONAL_TABLES`
    engine: <class 'sqlalchemy.engine.base.Engine'>
        User-defined database engine.
    limit: int
        Limit number of rows.
    additional_data: `list`
        Defaults to "export all available additional data" which is:
        `["unit_data", "eeg_data", "kwk_data", "permit_data"]`.
    chunksize: int or None
        Defines the chunksize of the tables export. Default to 500.000 which is roughly 2.5 GB.
    """

    renaming = column_renaming()

    unit_type_map_reversed = reverse_unit_type_map()

    with session_scope(engine=engine) as session:

        if tech:

            # Select orm tables for specified additional_data.
            orm_tables = {
                f"{dat}": getattr(orm, ORM_MAP[tech].get(dat, "KeyNotAvailable"), None)
                for dat in additional_data
            }

            # Filter for possible orm-additional_data combinations (not None)
            orm_tables = {k: v for k, v in orm_tables.items() if v is not None}

            # Build query based on available tables for tech and user input; always use basic units
            subtables = partially_suffixed_columns(
                orm.BasicUnit,
                renaming["basic_data"]["columns"],
                renaming["basic_data"]["suffix"],
            )

            # Extend table with columns from selected additional_data orm
            for addit_data_type, addit_data_orm in orm_tables.items():
                subtables.extend(
                    partially_suffixed_columns(
                        addit_data_orm,
                        renaming[addit_data_type]["columns"],
                        renaming[addit_data_type]["suffix"],
                    )
                )

            query_tech = Query(subtables, session=session)

            # Define joins based on available tables for data and user input
            if "unit_data" in orm_tables:
                query_tech = query_tech.join(
                    orm_tables["unit_data"],
                    orm.BasicUnit.EinheitMastrNummer
                    == orm_tables["unit_data"].EinheitMastrNummer,
                    isouter=True,
                )
            if "eeg_data" in orm_tables:
                query_tech = query_tech.join(
                    orm_tables["eeg_data"],
                    orm.BasicUnit.EegMastrNummer
                    == orm_tables["eeg_data"].EegMastrNummer,
                    isouter=True,
                )
            if "kwk_data" in orm_tables:
                query_tech = query_tech.join(
                    orm_tables["kwk_data"],
                    orm.BasicUnit.KwkMastrNummer
                    == orm_tables["kwk_data"].KwkMastrNummer,
                    isouter=True,
                )
            if "permit_data" in orm_tables:
                query_tech = query_tech.join(
                    orm_tables["permit_data"],
                    orm.BasicUnit.GenMastrNummer
                    == orm_tables["permit_data"].GenMastrNummer,
                    isouter=True,
                )

            # Restricted to technology
            query_tech = query_tech.filter(
                orm.BasicUnit.Einheittyp == unit_type_map_reversed[tech]
            )

            # Limit returned rows of query
            if limit:
                query_tech = query_tech.limit(limit)

            return query_tech

        if additional_table:

            orm_table = getattr(orm, ORM_MAP[additional_table], None)

            query_additional_tables = Query(orm_table, session=session)

            # Limit returned rows of query
            if limit:
                query_additional_tables = query_additional_tables.limit(limit)

            return query_additional_tables


def save_metadata(data: list = None, engine=None) -> None:
    """
    Save metadata during csv export.

    Parameters
    ----------
    data: list
        List of exported technologies for which metadata is needed.
    engine: <class 'sqlalchemy.engine.base.Engine'>
        User-defined database engine.

    Returns
    -------

    """
    data_path = get_data_version_dir()
    filenames = get_filenames()
    metadata_file = os.path.join(data_path, filenames["metadata"])
    unit_type_map_reversed = reverse_unit_type_map()

    with session_scope(engine=engine) as session:
        # check for latest db entry for exported technologies
        mastr_technologies = [unit_type_map_reversed[tech] for tech in data]
        newest_date = (
            session.query(orm.BasicUnit.DatumLetzteAktualisierung)
            .filter(orm.BasicUnit.Einheittyp.in_(mastr_technologies))
            .order_by(orm.BasicUnit.DatumLetzteAktualisierung.desc())
            .first()[0]
        )

    metadata = create_datapackage_meta_json(newest_date, data, json_serialize=False)

    with open(metadata_file, "w", encoding="utf-8") as f:
        json.dump(metadata, f, ensure_ascii=False, indent=4)

    log.info("Saved metadata")


def reverse_fill_basic_units(technology=None, engine=None):
    """
    The basic_units table is empty after bulk download.
    To enable csv export, the table is filled from extended
    tables reversely.

    .. warning::
    The basic_units table will be dropped and then recreated.
    Returns -------

    Parameters
    ----------
    technology: list of str
        Available technologies are in open_mastr.Mastr.to_csv()
    """

    with session_scope(engine=engine) as session:
        # Empty the basic_units table, because it will be filled entirely from extended tables
        session.query(getattr(orm, "BasicUnit", None)).delete()

        for tech in tqdm(technology, desc="Performing reverse fill of basic units: "):
            # Get the class of extended table
            unit_data_orm = getattr(orm, ORM_MAP[tech]["unit_data"], None)
            basic_unit_column_names = [
                column.name
                for column in getattr(orm, "BasicUnit", None).__mapper__.columns
            ]

            unit_columns_to_reverse_fill = [
                column
                for column in unit_data_orm.__mapper__.columns
                if column.name in basic_unit_column_names
            ]
            unit_column_names_to_reverse_fill = [
                column.name for column in unit_columns_to_reverse_fill
            ]

            unit_type_map_reversed = reverse_unit_type_map()

            # Add Einheittyp artificially
            unit_typ = "'" + unit_type_map_reversed.get(tech, None) + "'"
            unit_columns_to_reverse_fill.append(
                literal_column(unit_typ).label("Einheittyp")
            )
            unit_column_names_to_reverse_fill.append("Einheittyp")

            # Build query
            query = Query(unit_columns_to_reverse_fill, session=session)
            insert_query = insert(orm.BasicUnit).from_select(
                unit_column_names_to_reverse_fill, query
            )

            session.execute(insert_query)


def partially_suffixed_columns(mapper, column_names, suffix):
    """
    Add a suffix to a subset of ORM map tables for a query

    Parameters
    ----------
    mapper:
        SQLAlchemy ORM table mapper
    column_names: list
        Names of columns to be suffixed
    suffix: str
        Suffix that is append like + "_" + suffix

    Returns
    -------
    list
        List of ORM table mapper instance
    """
    columns = list(mapper.__mapper__.columns)
    return [
        _.label(f"{_.name}_{suffix}") if _.name in column_names else _ for _ in columns
    ]


def db_query_to_csv(db_query, data_table: str, chunksize: int) -> None:
    """
    Export database query to CSV file

    Save CSV files to the respective data version directory, see
    :meth:`open_mastr.utils.config.get_data_version_dir`.

    Parameters
    ----------
    db_query: <class 'sqlalchemy.orm.query.Query'>
        QueryORM-level SQL construction object that holds tables for export.
    data_table: str
        See list of available technologies or additional tables in
        `open_mastr.utils.constants.TECHNOLOGIES` and
        `open_mastr.utils.constants.ADDITIONAL_TABLES`
    chunksize: int
        Defines the size of the chunks that are saved to csv.
        Useful when export fails due to memory issues.
    """
    data_path = get_data_version_dir()
    filenames = get_filenames()

    # Set export settings per table type
    if data_table in TECHNOLOGIES:
        index = True
        index_col = "EinheitMastrNummer"
        index_label = "EinheitMastrNummer"
        csv_file = os.path.join(data_path, filenames["raw"][data_table]["joined"])
    if data_table in ADDITIONAL_TABLES:
        index = False
        index_col = None
        index_label = None
        csv_file = os.path.join(
            data_path, filenames["raw"]["additional_table"][data_table]
        )

    with db_query.session.bind.connect() as con:
        with con.begin():
            # Read data into pandas.DataFrame in chunks of max. 500000 rows of ~2.5 GB RAM
            for chunk_number, chunk_df in enumerate(
                pd.read_sql(
                    sql=db_query.statement,
                    con=con,
                    index_col=index_col,
                    chunksize=chunksize,
                )
            ):
                # For debugging purposes, check RAM usage of chunk_df
                # chunk_df.info(memory_usage='deep')

                # Make sure no duplicate column names exist
                assert not any(chunk_df.columns.duplicated())

                # Remove newline statements from certain strings
                if data_table in TECHNOLOGIES:
                    for col in ["Aktenzeichen", "Behoerde"]:
                        chunk_df[col] = chunk_df[col].str.replace("\r", "")

                if not chunk_df.empty:

                    if chunk_number == 0:
                        chunk_df.to_csv(
                            csv_file,
                            index=index,
                            index_label=index_label,
                            encoding="utf-8",
                        )
                        log.info(f"Created csv: {csv_file.split('/')[-1:]} ")
                    else:
                        chunk_df.to_csv(
                            csv_file,
                            mode="a",
                            header=False,
                            index=index,
                            index_label=index_label,
                            encoding="utf-8",
                        )
                        log.info(
                            f"Appended {len(chunk_df)} rows to: {csv_file.split('/')[-1:]}"
                        )


def rename_table(table, columns, engine) -> None:
    """
    Rename table based on translation dictionary.
    """
    alter_statements = []

    for column in columns:
        column = column["name"]

        if column in TRANSLATIONS:
            alter_statement = text(
                f"ALTER TABLE {table} RENAME COLUMN {column} TO {TRANSLATIONS[column]}"
            )
            alter_statements.append(alter_statement)

    with engine.connect() as connection:
        for statement in alter_statements:
            try:
                connection.execute(statement)
            except sqlalchemy.exc.OperationalError:
                continue


def create_translated_database_engine(engine, folder_path) -> sqlalchemy.engine.Engine:
    """
    Check if translated version of the database, as defined with engine parameter, exists.
    Return sqlite engine connected with the translated database.
    """

    if engine == "sqlite":
        db_path = os.path.join(folder_path, "open-mastr-translated.db")
    else:
        if "sqlite" not in engine.dialect.name:
            raise ValueError("engine has to be of type 'sqlite'")

        prev_path = r"{}".format(engine.url.database)
        engine.dispose()
        db_path = prev_path[:-3] + "-translated.db"

    if not os.path.exists(db_path):
        raise FileNotFoundError(
            f"no database at {db_path} found.\n"
            "make sure the database has been translated before with translate()"
        )

    return create_engine(f"sqlite:///{db_path}")



# import datetime
# import multiprocessing
# import logging
# import tqdm
#
# from open_mastr.soap_api.utils import is_time_blacklisted
# log = logging.getLogger(__name__)
#
# last_successful_download = datetime.datetime.now()
#
# # Check if worker threads need to be killed
#
#
# def _stop_execution(time_blacklist, timeout):
#     # Last successful execution was more than 10 minutes ago. Server seems
#     # unresponsive, so stop execution permanently by raising an error
#     if last_successful_download + datetime.timedelta(minutes=timeout) < datetime.datetime.now():
#         log.error('No response from server in the last {} minutes. Stopping execution.'.format(timeout))
#         raise ConnectionAbortedError
#     # Stop execution smoothly if current system time is in blacklist by
#     # returning. Calling function can decide to continue running later if
#     # needed
#     if time_blacklist and is_time_blacklisted(last_successful_download.time()):
#         log.info('Current time is in blacklist. Halting.')
#         return True
#     # ... Add more checks here if needed ...
#     return False
#



from utils.data_io import zenodo_upload, read_csv_data
from open_mastr.postprocessing.cleaning import cleaned_data
from open_mastr.postprocessing.postprocessing import postprocess
from open_mastr.postprocessing.postprocessing import to_csv

cleaned = cleaned_data()
zenodo_upload(data_stages=["raw", "cleaned"])
cleaned = read_csv_data("cleaned")
postprocess(cleaned)
to_csv()



from open_mastr.soap_api.mirror import MaStRMirror
import datetime

# Dump data
now = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
dump_file = f"{now}_open-mastr-mirror.backup"

mastr_refl = MaStRMirror()
mastr_refl.dump(dump_file)



from open_mastr.soap_api.mirror import MaStRMirror
import datetime

limit = None
technology = [
    "wind",
    "biomass",
    "combustion",
    "gsgk",
    "hydro",
    "nuclear",
    "storage",
    "solar",
]
data_types = ["unit_data", "eeg_data", "kwk_data", "permit_data"]
location_types = [
    "location_elec_generation",
    "location_elec_consumption",
    "location_gas_generation",
    "location_gas_consumption",
]
processes = 12

mastr_mirror = MaStRMirror(
    empty_schema=False, parallel_processes=processes, restore_dump=None
)

# Download basic unit data
mastr_mirror.backfill_basic(technology, limit=limit, date="latest")

# Download additional unit data
for tech in technology:
    # mastr_mirror.create_additional_data_requests(tech)
    for data_type in data_types:
        mastr_mirror.retrieve_additional_data(
            tech, data_type, chunksize=1000, limit=limit
        )

# Download basic location data
mastr_mirror.backfill_locations_basic(limit=limit, date="latest")

# Download extended location data
for location_type in location_types:
    mastr_mirror.retrieve_additional_location_data(location_type, limit=limit)



# __copyright__ = "© Reiner Lemoine Institut"
# __license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
# __url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
# __author__ = "Ludee; Bachibouzouk; solar-c"
# __issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
# __version__ = "v0.10.0"
#
# import logging
# import datetime
#
#
# log = logging.getLogger(__name__)
#
#
# def is_time_blacklisted(time):
#    times_blacklist = [
#        ('8:00', '18:00'),  # BNetzA Business hours
#        ('23:30', '00:10'),  # Daily database cronjob
#        # Add more if needed...
#    ]
#
#    # check if time is in a given interval between upper and lower
#    def in_interval(lower, upper):
#        # Convert str to datatime object
#        def parse_time(t):
#            return datetime.datetime.strptime(t, "%H:%M").time()
#        lower = parse_time(lower)
#        upper = parse_time(upper)
#
#        # Handle interval that spans over midnight (i.e. 23:30-0:30)
#        if lower > upper:
#            return (time <= upper or time >= lower)
#        # Handle all other intevals
#        return (lower <= time and upper >= time)
#
#    # check if time is in interval for each interval in the blacklist
#    in_interval = [in_interval(lower, upper) for lower, upper in times_blacklist]
#    return any(in_interval)



import time
from open_mastr.xml_download.utils_download_bulk import gen_url


def test_gen_url():
    when = time.strptime("2024-01-01", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20240101_23.2.zip"
    )

    when = time.strptime("2024-04-01", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20240401_23.2.zip"
    )

    when = time.strptime("2024-04-02", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20240402_24.1.zip"
    )

    when = time.strptime("2024-10-01", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20241001_24.1.zip"
    )

    when = time.strptime("2024-10-02", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20241002_24.2.zip"
    )

    when = time.strptime("2024-12-31", "%Y-%m-%d")
    url = gen_url(when)
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20241231_24.2.zip"
    )

    # Tests for use_version parameter

    when = time.strptime("2024-12-31", "%Y-%m-%d")
    url = gen_url(when, use_version="before")
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20241231_24.1.zip"
    )

    when = time.strptime("2024-12-31", "%Y-%m-%d")
    url = gen_url(when, use_version="after")
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20241231_25.1.zip"
    )

    when = time.strptime("2024-04-02", "%Y-%m-%d")
    url = gen_url(when, use_version="before")
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20240402_23.2.zip"
    )

    when = time.strptime("2024-04-02", "%Y-%m-%d")
    url = gen_url(when, use_version="after")
    assert type(url) == str
    assert (
        url
        == "https://download.marktstammdatenregister.de/Gesamtdatenexport_20240402_24.2.zip"
    )



import os
from sqlalchemy import inspect, create_engine

# import xml dependencies
from open_mastr.xml_download.utils_download_bulk import download_xml_Mastr
from open_mastr.xml_download.utils_write_to_database import (
    write_mastr_xml_to_database,
)

# import soap_API dependencies
from open_mastr.soap_api.mirror import MaStRMirror

from open_mastr.utils.helpers import (
    print_api_settings,
    validate_api_credentials,
    validate_parameter_format_for_download_method,
    validate_parameter_format_for_mastr_init,
    validate_parameter_data,
    transform_data_parameter,
    parse_date_string,
    transform_date_parameter,
    data_to_include_tables,
    create_db_query,
    db_query_to_csv,
    reverse_fill_basic_units,
)
from open_mastr.utils.config import (
    create_data_dir,
    get_data_version_dir,
    get_project_home_dir,
    get_output_dir,
    setup_logger,
)
import open_mastr.utils.orm as orm

# import initialize_database dependencies
from open_mastr.utils.helpers import (
    create_database_engine,
    rename_table,
    create_translated_database_engine,
)

# constants
from open_mastr.utils.constants import TECHNOLOGIES, ADDITIONAL_TABLES

# setup logger
log = setup_logger()


class Mastr:
    """
    `Mastr` is used to download the MaStR database and keep it up-to-date.

    A SQL database is used to mirror the MaStR database. It can be filled with
    data either from the MaStR-bulk download or from the MaStR-API.

    !!! example

        ```python
        from open_mastr import Mastr

        db = Mastr()
        db.download()
        ```

    Parameters
    ----------
    engine : {'sqlite', sqlalchemy.engine.Engine}, optional
        Defines the engine of the database where the MaStR is mirrored to.
        Default is 'sqlite'.
    connect_to_translated_db: boolean, optional
            Allows connection to an existing translated database. Default is 'False'.
            Only for 'sqlite'-type engines.



    """

    def __init__(self, engine="sqlite", connect_to_translated_db=False) -> None:
        validate_parameter_format_for_mastr_init(engine)

        self.output_dir = get_output_dir()
        self.home_directory = get_project_home_dir()
        self._sqlite_folder_path = os.path.join(self.output_dir, "data", "sqlite")
        os.makedirs(self._sqlite_folder_path, exist_ok=True)

        self.is_translated = connect_to_translated_db
        if connect_to_translated_db:
            self.engine = create_translated_database_engine(
                engine, self._sqlite_folder_path
            )
        else:
            self.engine = create_database_engine(engine, self._sqlite_folder_path)

        print(
            f"Data will be written to the following database: {self.engine.url}\n"
            "If you run into problems, try to "
            "delete the database and update the package by running "
            "'pip install --upgrade open-mastr'\n"
        )

        orm.Base.metadata.create_all(self.engine)

    def download(
        self,
        method="bulk",
        data=None,
        date=None,
        bulk_cleansing=True,
        api_processes=None,
        api_limit=50,
        api_chunksize=1000,
        api_data_types=None,
        api_location_types=None,
        **kwargs,
    ) -> None:
        """
        Download the MaStR either via the bulk download or via the MaStR API and write it to a
        SQLite database.

        Parameters
        ----------
        method : 'API' or 'bulk', optional
            Either "API" or "bulk". Determines whether the data is downloaded via the
            zipped bulk download or via the MaStR API. The latter requires an account
            from marktstammdatenregister.de,
            (see :ref:`Configuration <Configuration>`). Default to 'bulk'.
        data : str or list or None, optional
            Determines which types of data are written to the database. If None, all data is
            used. If it is a list, possible entries are listed below with respect to the download method. Missing categories are
            being developed. If only one data is of interest, this can be given as a string. Default to None, where all data is included.

            | Data                  | Bulk | API  |
            |-----------------------|------|------|
            | "wind"                | Yes  | Yes  |
            | "solar"               | Yes  | Yes  |
            | "biomass"             | Yes  | Yes  |
            | "hydro"               | Yes  | Yes  |
            | "gsgk"                | Yes  | Yes  |
            | "combustion"          | Yes  | Yes  |
            | "nuclear"             | Yes  | Yes  |
            | "gas"                 | Yes  | Yes  |
            | "storage"             | Yes  | Yes  |
            | "storage_units"       | Yes  | Yes  |
            | "electricity_consumer"| Yes  | No   |
            | "location"            | Yes  | Yes  |
            | "market"              | Yes  | No   |
            | "grid"                | Yes  | No   |
            | "balancing_area"      | Yes  | No   |
            | "permit"              | Yes  | Yes  |
            | "deleted_units"       | Yes  | No   |
            | "deleted_market_actors"| Yes | No   |
            | "retrofit_units"      | Yes  | No   |
        date : None or `datetime.datetime` or str, optional

            | date                  | Bulk | API  |
            |-----------------------|------|------|
            | "today"                | latest files are downloaded from marktstammdatenregister.de  | -  |
            | "20230101"      | If file from this date exists locally, it is used. Otherwise it throws an error (You can only receive todays data from the server)  | -   |
            | "existing"               | Use latest downloaded zipped xml files, throws an error if the bulk download folder is empty  | -  |
            | "latest"               | -  | Retrieve data that is newer than the newest data already in the table  |
            | datetime.datetime(2020, 11, 27)      | -  | Retrieve data that is newer than this time stamp   |
            | None      | set date="today"  | set date="latest"   |

            Default to `None`.
        bulk_cleansing : bool, optional
            If set to True, data cleansing is applied after the download (which is recommended).
            In its original format, many entries in the MaStR are encoded with IDs. Columns like
            `state` or `fueltype` do not contain entries such as "Hessen" or "Braunkohle", but instead
            only contain IDs. Cleansing replaces these IDs with their corresponding original entries.
        api_processes : int or None or "max", optional
            Number of parallel processes used to download additional data.
            Defaults to `None`. If set to "max", the maximum number of possible processes
            is used.

            !!! warning

                The implementation of parallel processes is currently under construction.
                Please let the argument `api_processes` at the default value `None`.
        api_limit : int or None, optional
            Limit the number of units that data is downloaded for. Defaults to `None` which refers
            to query data for existing data requests, for example created by
            [`create_additional_data_requests`][open_mastr.soap_api.mirror.MaStRMirror.create_additional_data_requests]. Note: There is a limited number of
            requests you are allowed to have per day, so setting api_limit to a value is
            recommended.
        api_chunksize : int or None, optional
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        api_data_types : list or None, optional
            Select the type of additional data that should be retrieved. Choose from
            "unit_data", "eeg_data", "kwk_data", "permit_data". Defaults to all.
        api_location_types : list or None, optional
            Select the type of location that should be retrieved. Choose from
            "location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption". Defaults to all.
        """

        if self.is_translated:
            raise TypeError(
                "You are currently connected to a translated database.\n"
                "A translated database cannot be further processed."
            )

        validate_parameter_format_for_download_method(
            method=method,
            data=data,
            date=date,
            bulk_cleansing=bulk_cleansing,
            api_processes=api_processes,
            api_limit=api_limit,
            api_chunksize=api_chunksize,
            api_data_types=api_data_types,
            api_location_types=api_location_types,
            **kwargs,
        )
        (
            data,
            api_data_types,
            api_location_types,
            harm_log,
        ) = transform_data_parameter(
            method, data, api_data_types, api_location_types, **kwargs
        )

        date = transform_date_parameter(self, method, date, **kwargs)

        if method == "bulk":
            # Find the name of the zipped xml folder
            bulk_download_date = parse_date_string(date)
            xml_folder_path = os.path.join(self.output_dir, "data", "xml_download")
            os.makedirs(xml_folder_path, exist_ok=True)
            zipped_xml_file_path = os.path.join(
                xml_folder_path,
                f"Gesamtdatenexport_{bulk_download_date}.zip",
            )
            download_xml_Mastr(zipped_xml_file_path, date, xml_folder_path)

            print(
                f"\nWould you like to speed up the bulk download?\n"
                f"Try our new parallelized processing by setting os.environ['USE_RECOMMENDED_NUMBER_OF_PROCESSES'] = True "
                f"or configure your own number of processes via os.environ['NUMBER_OF_PROCESSES'] = your_number\n"
            )

            write_mastr_xml_to_database(
                engine=self.engine,
                zipped_xml_file_path=zipped_xml_file_path,
                data=data,
                bulk_cleansing=bulk_cleansing,
                bulk_download_date=bulk_download_date,
            )

        if method == "API":
            validate_api_credentials()

            # Set api_processes to None in order to avoid the malfunctioning usage
            if api_processes:
                api_processes = None
                print(
                    "Warning: The implementation of parallel processes "
                    "is currently under construction. Please let "
                    "the argument api_processes at the default value None."
                )

            print_api_settings(
                harmonisation_log=harm_log,
                data=data,
                date=date,
                api_data_types=api_data_types,
                api_chunksize=api_chunksize,
                api_limit=api_limit,
                api_processes=api_processes,
                api_location_types=api_location_types,
            )

            mastr_mirror = MaStRMirror(
                engine=self.engine,
                parallel_processes=api_processes,
                restore_dump=None,
            )
            # Download basic unit data
            mastr_mirror.backfill_basic(data, limit=api_limit, date=date)

            # Download additional unit data
            for tech in data:
                # mastr_mirror.create_additional_data_requests(data)
                for data_type in api_data_types:
                    mastr_mirror.retrieve_additional_data(
                        tech, data_type, chunksize=api_chunksize, limit=api_limit
                    )

            # Download basic location data
            mastr_mirror.backfill_locations_basic(limit=api_limit, date="latest")

            # Download extended location data
            if api_location_types:
                for location_type in api_location_types:
                    mastr_mirror.retrieve_additional_location_data(
                        location_type, limit=api_limit
                    )

    def to_csv(
        self, tables: list = None, chunksize: int = 500000, limit: int = None
    ) -> None:
        """
        Save the database as csv files along with the metadata file.
        If 'tables=None' all possible tables will be exported.

        Parameters
        ------------
        tables: None or list
            For exporting selected tables choose from:
                ["wind", "solar", "biomass", "hydro", "gsgk", "combustion", "nuclear", "storage",
                "balancing_area", "electricity_consumer", "gas_consumer", "gas_producer",
                "gas_storage", "gas_storage_extended",
                "grid_connections", "grids", "market_actors", "market_roles",
                "locations_extended", "permit", "deleted_units", "storage_units"]
        chunksize: int
            Defines the chunksize of the tables export.
            Default value is 500.000 rows to include in each chunk.
        limit: None or int
            Limits the number of exported data rows.
        """

        if self.is_translated:
            raise TypeError(
                "You are currently connected to a translated database.\n"
                "A translated database cannot be used for the csv export."
            )

        log.info("Starting csv-export")

        data_path = get_data_version_dir()

        create_data_dir()

        # Validate and parse tables parameter
        validate_parameter_data(method="csv_export", data=tables)
        (
            data,
            api_data_types,
            api_location_types,
            harm_log,
        ) = transform_data_parameter(
            method="bulk", data=tables, api_data_types=None, api_location_types=None
        )

        # Determine tables to export
        technologies_to_export = []
        additional_tables_to_export = []
        for table in data:
            if table in TECHNOLOGIES:
                technologies_to_export.append(table)
            elif table in ADDITIONAL_TABLES:
                additional_tables_to_export.append(table)
            else:
                additional_tables_to_export.extend(
                    data_to_include_tables([table], mapping="export_db_tables")
                )

        if technologies_to_export:
            log.info(f"Technology tables: {technologies_to_export}")
        if additional_tables_to_export:
            log.info(f"Additional tables: {additional_tables_to_export}")

        log.info(f"Tables are saved to: {data_path}")

        reverse_fill_basic_units(technology=technologies_to_export, engine=self.engine)

        # Export technologies to csv
        for tech in technologies_to_export:
            db_query_to_csv(
                db_query=create_db_query(tech=tech, limit=limit, engine=self.engine),
                data_table=tech,
                chunksize=chunksize,
            )
        # Export additional tables to csv
        for addit_table in additional_tables_to_export:
            db_query_to_csv(
                db_query=create_db_query(
                    additional_table=addit_table, limit=limit, engine=self.engine
                ),
                data_table=addit_table,
                chunksize=chunksize,
            )

        # FIXME: Currently metadata is only created for technology data, Fix in #386
        # Configure and save data package metadata file along with data
        # save_metadata(data=technologies_to_export, engine=self.engine)

    def translate(self) -> None:
        """
        A database can be translated only once.

        Deletes translated versions of the currently connected database.

        Translates currently connected database,renames it with '-translated'
        suffix and updates self.engine's path accordingly.

        !!! example
            ```python

            from open_mastr import Mastr
            import pandas as pd

            db = Mastr()
            db.download(data='biomass')
            db.translate()

            df = pd.read_sql(sql='biomass_extended', con=db.engine)
            print(df.head(10))
            ```

        """

        if "sqlite" not in self.engine.dialect.name:
            raise ValueError("engine has to be of type 'sqlite'")
        if self.is_translated:
            raise TypeError("The currently connected database is already translated.")

        inspector = inspect(self.engine)
        old_path = r"{}".format(self.engine.url.database)
        new_path = old_path[:-3] + "-translated.db"

        if os.path.exists(new_path):
            try:
                os.remove(new_path)
            except Exception as e:
                print(f"An error occurred: {e}")

            print("Replacing previous version of the translated database...")

        for table in inspector.get_table_names():
            rename_table(table, inspector.get_columns(table), self.engine)

        self.engine.dispose()

        try:
            os.rename(old_path, new_path)
            print(f"Database '{old_path}' changed to '{new_path}'")
        except Exception as e:
            print(f"An error occurred: {e}")

        self.engine = create_engine(f"sqlite:///{new_path}")
        self.is_translated = True



from geoalchemy2 import Geometry
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData
from sqlalchemy import (
    Column,
    Integer,
    String,
    Float,
    Sequence,
    DateTime,
    Boolean,
    func,
    Date,
    JSON,
)
from sqlalchemy.dialects.postgresql import JSONB

cleaned_schema = "model_draft"
meta = MetaData(schema=cleaned_schema)
Base = declarative_base(metadata=meta)


class BasicUnit(object):

    EinheitMastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Name = Column(String)
    Einheitart = Column(String)
    Einheittyp = Column(String)
    Standort = Column(String)
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String)
    Anlagenbetreiber = Column(String)
    EegMastrNummer = Column(String)
    KwkMastrNummer = Column(String)
    SpeMastrNummer = Column(String)
    GenMastrNummer = Column(String)
    BestandsanlageMastrNummer_basic = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(String)
    StatisikFlag_basic = Column(String)


class Extended(object):

    EinheitMastrNummer_extended = Column(String)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMastrNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(String)
    NetzbetreiberpruefungDatum = Column(DateTime(timezone=True))
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(String)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String)
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String)
    Ort = Column(String)
    Laengengrad = Column(String)
    Breitengrad = Column(String)
    UtmZonenwert = Column(String)
    UtmEast = Column(String)
    UtmNorth = Column(String)
    GaussKruegerHoch = Column(String)
    GaussKruegerRechts = Column(String)
    Meldedatum = Column(DateTime(timezone=True))
    GeplantesInbetriebnahmedatum = Column(DateTime(timezone=True))
    Inbetriebnahmedatum = Column(DateTime(timezone=True))
    DatumEndgueltigeStilllegung = Column(DateTime(timezone=True))
    DatumBeginnVoruebergehendeStilllegung = Column(DateTime(timezone=True))
    DatumWiederaufnahmeBetrieb = Column(DateTime(timezone=True))
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus_extended = Column(String)
    BestandsanlageMastrNummer = Column(String)
    NichtVorhandenInMigriertenEinheiten_extended = Column(Boolean)
    AltAnlagenbetreiberMastrNummer = Column(String)
    DatumDesBetreiberwechsels = Column(DateTime(timezone=True))
    DatumRegistrierungDesBetreiberwechsels = Column(DateTime(timezone=True))
    StatisikFlag = Column(String)
    NameStromerzeugungseinheit = Column(String)
    Weic = Column(String)
    WeicDisplayName = Column(String)
    Kraftwerksnummer = Column(String)
    Energietraeger = Column(String)
    Bruttoleistung_extended = Column(Float)
    Nettonennleistung = Column(Float)
    AnschlussAnHoechstOderHochSpannung = Column(Boolean)
    Schwarzstartfaehigkeit = Column(Boolean)
    Inselbetriebsfaehigkeit = Column(Boolean)
    Einsatzverantwortlicher = Column(String)
    FernsteuerbarkeitNb = Column(Boolean)
    FernsteuerbarkeitDv = Column(Boolean)
    FernsteuerbarkeitDr = Column(Boolean)
    Einspeisungsart = Column(String)
    PraequalifiziertFuerRegelenergie = Column(Boolean)
    GenMastrNummer_extended = Column(String)
    geom = Column(Geometry("POINT"))
    comment = Column(String)


class Eeg(object):

    EegMastrNummer_eeg = Column(String)
    Meldedatum_eeg = Column(Date)
    DatumLetzteAktualisierung_eeg = Column(DateTime(timezone=True))
    EegInbetriebnahmedatum = Column(Date)
    VerknuepfteEinheit = Column(String)


class WindEeg(Eeg):

    AnlagenkennzifferAnlagenregister = Column(String)
    AnlagenschluesselEeg = Column(String)
    PrototypAnlage = Column(Boolean)
    PilotAnlage = Column(Boolean)
    InstallierteLeistung = Column(Float)
    VerhaeltnisErtragsschaetzungReferenzertrag = Column(Float)
    VerhaeltnisReferenzertragErtrag5Jahre = Column(Float)
    VerhaeltnisReferenzertragErtrag10Jahre = Column(Float)
    VerhaeltnisReferenzertragErtrag15Jahre = Column(Float)
    AusschreibungZuschlag = Column(Boolean)
    Zuschlagsnummer = Column(String)
    AnlageBetriebsstatus = Column(String)


class SolarEeg(Eeg):

    InanspruchnahmeZahlungNachEeg = Column(Boolean)
    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    RegistrierungsnummerPvMeldeportal = Column(String)
    MieterstromZugeordnet = Column(Boolean)
    MieterstromMeldedatum = Column(Date)
    MieterstromErsteZuordnungZuschlag = Column(Date)
    AusschreibungZuschlag = Column(Boolean)
    ZugeordneteGebotsmenge = Column(Float)
    Zuschlagsnummer = Column(String)
    AnlageBetriebsstatus = Column(String)


class BiomassEeg(Eeg):

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AusschliesslicheVerwendungBiomasse = Column(Boolean)
    AusschreibungZuschlag = Column(Boolean)
    Zuschlagsnummer = Column(String)
    BiogasInanspruchnahmeFlexiPraemie = Column(Boolean)
    BiogasDatumInanspruchnahmeFlexiPraemie = Column(Date)
    BiogasLeistungserhoehung = Column(Boolean)
    BiogasDatumLeistungserhoehung = Column(Date)
    BiogasUmfangLeistungserhoehung = Column(Float)
    BiogasGaserzeugungskapazitaet = Column(Float)
    BiogasHoechstbemessungsleistung = Column(Float)
    BiomethanErstmaligerEinsatz = Column(Date)
    AnlageBetriebsstatus = Column(String)


class GsgkEeg(Eeg):

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AnlageBetriebsstatus = Column(String)


class HydroEeg(Eeg):

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AnlageBetriebsstatus = Column(String)
    Ertuechtigung = Column(JSON)


class StorageEeg(Eeg):
    pass


class Kwk(object):

    KwkMastrNummer_kwk = Column(String)
    AusschreibungZuschlag_kwk = Column(Boolean)
    Zuschlagnummer = Column(String)
    DatumLetzteAktualisierung_kwk = Column(DateTime(timezone=True))
    Inbetriebnahmedatum_kwk = Column(Date)
    Meldedatum_kwk = Column(Date)
    ThermischeNutzleistung = Column(Float)
    ElektrischeKwkLeistung = Column(Float)
    VerknuepfteEinheiten_kwk = Column(String)
    AnlageBetriebsstatus_kwk = Column(String)


class Permit(object):

    GenMastrNummer_permit = Column(String)
    DatumLetzteAktualisierung_permit = Column(DateTime(timezone=True))
    Art = Column(String)
    Datum = Column(Date)
    Behoerde = Column(String)
    Aktenzeichen = Column(String)
    Frist = Column(Date)
    WasserrechtsNummer = Column(String)
    WasserrechtAblaufdatum = Column(Date)
    Meldedatum_permit = Column(Date)
    VerknuepfteEinheiten = Column(String)


class WindCleaned(Permit, WindEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_wind_clean"

    # wind specific attributes
    NameWindpark = Column(String)
    Lage = Column(String)
    Seelage = Column(String)
    ClusterOstsee = Column(String)
    ClusterNordsee = Column(String)
    Hersteller = Column(String)
    HerstellerId = Column(String)
    Technologie = Column(String)
    Typenbezeichnung = Column(String)
    Nabenhoehe = Column(Float)
    Rotordurchmesser = Column(Float)
    Rotorblattenteisungssystem = Column(Boolean)
    AuflageAbschaltungLeistungsbegrenzung = Column(Boolean)
    AuflagenAbschaltungSchallimmissionsschutzNachts = Column(Boolean)
    AuflagenAbschaltungSchallimmissionsschutzTagsueber = Column(Boolean)
    AuflagenAbschaltungSchattenwurf = Column(Boolean)
    AuflagenAbschaltungTierschutz = Column(Boolean)
    AuflagenAbschaltungEiswurf = Column(Boolean)
    AuflagenAbschaltungSonstige = Column(Boolean)
    Wassertiefe = Column(Float)
    Kuestenentfernung = Column(Float)
    EegMastrNummer_extended = Column(String)
    tags = Column(JSONB)
    geom_3035 = Column(Geometry("POINT", srid=3035))


class SolarCleaned(Permit, SolarEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_solar_clean"

    zugeordneteWirkleistungWechselrichter = Column(Float)
    GemeinsamerWechselrichterMitSpeicher = Column(String)
    AnzahlModule = Column(Integer)
    Lage = Column(String)
    Leistungsbegrenzung = Column(String)
    EinheitlicheAusrichtungUndNeigungswinkel = Column(Boolean)
    Hauptausrichtung = Column(String)
    HauptausrichtungNeigungswinkel = Column(String)
    Nebenausrichtung = Column(String)
    NebenausrichtungNeigungswinkel = Column(String)
    InAnspruchGenommeneFlaeche = Column(String)
    ArtDerFlaeche = Column(String)
    InAnspruchGenommeneAckerflaeche = Column(Float)
    Nutzungsbereich = Column(String)
    EegMastrNummer_extended = Column(String)


class BiomassCleaned(Permit, Kwk, BiomassEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_biomass_clean"

    Hauptbrennstoff = Column(String)
    Biomasseart = Column(String)
    Technologie = Column(String)
    EegMastrNummer_extended = Column(String)
    KwkMastrNummer_extended = Column(String)


class CombustionCleaned(Permit, Kwk, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_combustion_clean"

    NameKraftwerk = Column(String)
    NameKraftwerksblock = Column(String)
    DatumBaubeginn = Column(Date)
    AnzeigeEinerStilllegung = Column(Boolean)
    ArtDerStilllegung = Column(String)
    DatumBeginnVorlaeufigenOderEndgueltigenStilllegung = Column(Date)
    SteigerungNettonennleistungKombibetrieb = Column(Float)
    AnlageIstImKombibetrieb = Column(Boolean)
    MastrNummernKombibetrieb = Column(String)
    NetzreserveAbDatum = Column(Date)
    SicherheitsbereitschaftAbDatum = Column(Date)
    Hauptbrennstoff = Column(String)
    WeitererHauptbrennstoff = Column(String)
    WeitereBrennstoffe = Column(String)
    VerknuepfteErzeugungseinheiten = Column(String)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    AnteiligNutzungsberechtigte = Column(String)
    Notstromaggregat = Column(Boolean)
    Einsatzort = Column(String)
    KwkMastrNummer_extended = Column(String)  # changed here
    Technologie = Column(String)


class GsgkCleaned(Permit, Kwk, GsgkEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_gsgk_clean"

    Technologie = Column(String)
    KwkMastrNummer_extended = Column(String)
    EegMastrNummer_extended = Column(String)


class HydroCleaned(Permit, HydroEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_hydro_clean"

    NameKraftwerk = Column(String)
    ArtDerWasserkraftanlage = Column(String)
    AnzeigeEinerStilllegung = Column(Boolean)
    ArtDerStilllegung = Column(String)
    DatumBeginnVorlaeufigenOderEndgueltigenStilllegung = Column(Date)
    MinderungStromerzeugung = Column(Boolean)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    ArtDesZuflusses = Column(String)
    EegMastrNummer_extended = Column(String)


class NuclearCleaned(Permit, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_nuclear_clean"

    NameKraftwerk = Column(String)
    NameKraftwerksblock = Column(String)
    Technologie = Column(String)


class StorageCleaned(Permit, StorageEeg, Extended, BasicUnit, Base):
    __tablename__ = "bnetza_mastr_storage_clean"

    Einsatzort = Column(String)
    AcDcKoppelung = Column(String)
    Batterietechnologie = Column(String)
    PumpbetriebLeistungsaufnahme = Column(Float)
    PumpbetriebKontinuierlichRegelbar = Column(Boolean)
    Pumpspeichertechnologie = Column(String)
    Notstromaggregat = Column(Boolean)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    ZugeordnenteWirkleistungWechselrichter = Column(Float)
    NutzbareSpeicherkapazitaet = Column(Float)
    SpeMastrNummer_extended = Column(String)
    EegMastrNummer_extended = Column(String)
    EegAnlagentyp = Column(String)
    Technologie = Column(String)



import datetime
import json
import pandas as pd
from sqlalchemy import text, String
from geoalchemy2 import Geometry, WKTElement
import numpy as np
import os
from urllib.request import urlretrieve
from open_mastr.postprocessing import orm
from open_mastr.utils.config import setup_logger, get_filenames, get_data_version_dir
from open_mastr.soap_api.metadata.create import create_datapackage_meta_json
from open_mastr.utils.helpers import chunks, session_scope, db_engine
import geopandas as gpd
from shapely.wkb import loads as wkb_loads


log = setup_logger()

BKG_VG250 = {"schema": "boundaries", "table": "bkg_vg250_1_sta_union_mview"}

OSM_PLZ = {"schema": "boundaries", "table": "osm_postcode"}

OFFSHORE = {"schema": "model_draft", "table": "rli_boundaries_offshore"}

OSM_WINDPOWER = {"schema": "model_draft", "table": "mastr_osm_deu_point_windpower"}

OEP_QUERY_PATTERN = "https://openenergy-platform.org/api/v0/schema/{schema}/tables/{table}/rows?form=csv"

DATA_BASE_PATH = "data"

MASTR_RAW_SCHEMA = "model_draft"
OPEN_MASTR_SCHEMA = "model_draft"

TECHNOLOGIES = [
    "wind",
    "hydro",
    "solar",
    "biomass",
    "combustion",
    "nuclear",
    "gsgk",
    "storage",
]

orm_map = {
    "wind": {
        "cleaned": "WindCleaned",
    },
    "solar": {
        "cleaned": "SolarCleaned",
    },
    "biomass": {
        "cleaned": "BiomassCleaned",
    },
    "combustion": {
        "cleaned": "CombustionCleaned",
    },
    "gsgk": {
        "cleaned": "GsgkCleaned",
    },
    "hydro": {
        "cleaned": "HydroCleaned",
    },
    "nuclear": {
        "cleaned": "NuclearCleaned",
    },
    "storage": {
        "cleaned": "StorageCleaned",
    },
}


def wkb_hexer(line):
    """
    Convert WKB to hex format

    Parameters
    ----------
    line : shapely
        Shapely Well Known Binary (WKB) representation

    Returns
    -------
    shapely.wkb_hex
        Hex representation of geo data

    """
    return line.wkb_hex


def table_to_db(csv_data, table, schema, conn, geom_col="geom", srid=4326):
    """
    Import data table into PostgreSQL database

    Parameters
    ----------
    csv_data : pandas.DataFrame
        The table data
    table : str
        Table name
    schema : str
        Schema name
    conn : sqlalchemy.engine.Connection
        Database connection
    geom_col : str
        Name of column which is expected to hold geom data (defaults to 'geom')
    """

    # Create schemas
    query = "CREATE SCHEMA IF NOT EXISTS {schema}".format(schema=schema)
    conn.execute(query)

    csv_data.to_sql(
        table,
        con=conn,
        schema=schema,
        dtype={
            geom_col: Geometry(srid=srid),
            "plz": String(),
        },
        chunksize=100000,
        if_exists="replace",
    )


def table_to_db_orm(mapper, data, chunksize=10000):
    """
    Import data table into PostgreSQL database using ORM bulk insert

    Parameters
    ----------
    mapper: SQLAlchemy decarative mapping
        Table ORM
    data: pandas.DataFrame
        Tabular data for one technology
    chunksize: int, optional
        Size of data chunks for database insertation. Data gets inserted in chunks of size `chunksize`.
    """

    with session_scope() as session:
        for chunk in chunks(data.reset_index().to_dict(orient="records"), chunksize):
            session.bulk_insert_mappings(mapper, chunk)
            # Commit each chunk separately
            session.commit()


def import_boundary_data_csv(schema, table, index_col="id", srid=4326):
    """
    Import additional data for post-processing

    Parameters
    ----------
    schema : str
        Schema of `table`
    table : str
        Table name
    index_col : str
        Column used as index (defaults to 'id')
    """

    csv_file = "{schema}_{table}.csv".format(schema=schema, table=table)

    # Check if file is locally available
    csv_file_exists = os.path.isfile(csv_file)

    with db_engine().connect() as con:

        # Check if table already exists
        table_query = "SELECT to_regclass('{schema}.{table}');".format(
            schema=schema, table=table
        )
        table_name = "{schema}.{table}".format(schema=schema, table=table)
        table_exists = table_name in con.execute(table_query).first().values()

        if not table_exists:
            # Download CSV file if it does not exist
            if not csv_file_exists:
                log.info(
                    "Downloading table {schema}.{table} from OEP".format(
                        schema=schema, table=table
                    )
                )
                urlretrieve(
                    OEP_QUERY_PATTERN.format(schema=schema, table=table), csv_file
                )
            else:
                log.info("Found {} locally.".format(csv_file))

            # Read CSV file
            csv_data = pd.read_csv(csv_file, index_col=index_col)

            # Prepare geom data for DB upload
            csv_data["geom"] = csv_data["geom"].apply(
                lambda x: WKTElement(wkb_loads(x, hex=True).wkt, srid=srid)
            )

            # Insert to db
            table_to_db(csv_data, table, schema, con, srid=srid)
            log.info("Data from {} successfully imported to database.".format(csv_file))
        else:
            log.info(
                "Table '{schema}.{table}' already exists in local database".format(
                    schema=schema, table=table
                )
            )


def add_geom_col(df, lat_col="Breitengrad", lon_col="Laengengrad", srid=4326):
    """
    Creates a geometry column based on lat/long

    Parameters
    ----------
    df : pandas.DataFrame
        Data
    lat_col : str
        Column name with lat coordinate
    lon_col : str
        Column name with long coordinate
    srid : int
        Defines the spatial reference system of geo data (defaults to 4326)

    Returns
    -------
    GeoDataFrame
        Read MaStR raw data with added geom column
    """

    # Split data into with and without coordinates
    df_with_coords = df.loc[~(df["Breitengrad"].isna() | df["Laengengrad"].isna())]

    # Just select data with lat/lon in range [(-90,90), (-180,180)]
    df_with_coords = df_with_coords[
        ~(
            (df_with_coords["Breitengrad"] < -90)
            | (df_with_coords["Breitengrad"] > 90)
            | (df_with_coords["Laengengrad"] < -180)
            | (df_with_coords["Laengengrad"] > 180)
        )
    ]
    df_no_coords = df.loc[~df.index.isin(df_with_coords.index)]

    gdf = gpd.GeoDataFrame(
        df_with_coords,
        geometry=gpd.points_from_xy(df_with_coords[lon_col], df_with_coords[lat_col]),
        crs="EPSG:{}".format(srid),
    )
    gdf["geom"] = gdf["geometry"].apply(lambda x: WKTElement(x.wkt, srid=srid))
    gdf.drop(columns=["geometry"], inplace=True)

    gdf = gdf.append(df_no_coords)

    return gdf


def import_bnetz_mastr_csv(mastr_cleaned):
    """
    Import MaStR raw data and create geom column

    Parameters
    ----------
    mastr_cleaned: dict of pandas.DataFrame
        Cleaned MaStR data in a dictionary of dataframes keyed by technology.
    """

    for k, d in mastr_cleaned.items():

        # Create 'geom' column from lat/lon
        gdf = add_geom_col(d)

        # Import to local database
        log.info(f"Import data to database for {k}")
        mapper = getattr(orm, orm_map[k]["cleaned"])
        table_to_db_orm(mapper, gdf.replace({np.nan: None}), chunksize=100000)
        log.info(f"Data for {k} successfully imported to database.")


def run_sql_postprocessing():
    """
    Execute SQL scripts ins `db-cleansing/`
    """

    with db_engine().connect().execution_options(autocommit=True) as con:

        for tech_name in TECHNOLOGIES:
            if tech_name not in ["gsgk", "storage", "nuclear"]:
                log.info(f"Run post-processing on {tech_name} data")
                # Read SQL query from file
                with open(
                    os.path.join(
                        os.path.dirname(__file__),
                        "db-cleansing",
                        "rli-mastr-{tech_name}-cleansing.sql".format(
                            tech_name=tech_name
                        ),
                    )
                ) as file:
                    escaped_sql = text(file.read())

                # Execute query
                con.execute(escaped_sql)


def postprocess(mastr_cleaned):
    """
    Run post-processing

    Import cleaned MaStR data to PostgreSQL database, retrieve additional data for post-processing, enrich, and
    prepare data for further analysis.

    Parameters
    ----------
    mastr_cleaned: dict of pandas.DataFrame
        Cleaned MaStR data in a dictionary of dataframes keyed by technology.
    """
    # Create cleaned tables
    engine = db_engine()
    with engine.connect().execution_options(autocommit=True) as con:
        con.execute(f"DROP SCHEMA IF EXISTS {orm.Base.metadata.schema} CASCADE;")
        con.execute(f"CREATE SCHEMA {orm.Base.metadata.schema};")
    orm.Base.metadata.create_all(engine)

    # Import external data (most boundaries)
    import_boundary_data_csv(BKG_VG250["schema"], BKG_VG250["table"], srid=3035)
    import_boundary_data_csv(OSM_PLZ["schema"], OSM_PLZ["table"])
    import_boundary_data_csv(OFFSHORE["schema"], OFFSHORE["table"])
    import_boundary_data_csv(OSM_WINDPOWER["schema"], OSM_WINDPOWER["table"])

    # Import MaStR raw data in o database
    import_bnetz_mastr_csv(mastr_cleaned)

    # Process data
    run_sql_postprocessing()


def to_csv(limit=None):
    """
    Write post-processed MaStR data to CSV files

    Write structurally similar data as :meth:`~.open_mastr.soap_api.mirror.MaStRMirror.to_csv`, but with additional
    columns and post-processed data.

    Metadata gets created decribing post-processed and raw data as one datapackage.

    Parameters
    ----------
    limit: int, optional
        Limit number of rows exported to CSV file. This applied independently for each file/technology.
    """
    data_path = get_data_version_dir()
    filenames = get_filenames()
    newest_date = datetime.datetime(1900, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

    for tech in TECHNOLOGIES:
        with session_scope() as session:
            orm_tech = getattr(orm, orm_map[tech]["cleaned"])
            query = session.query(orm_tech).limit(limit)
            df = pd.read_sql(
                query.statement, query.session.bind, index_col="EinheitMastrNummer"
            )

        csv_file = os.path.join(data_path, filenames["postprocessed"][tech])

        df.to_csv(
            csv_file, index=True, index_label="EinheitMastrNummer", encoding="utf-8"
        )

        if df["DatumLetzteAktualisierung"].max() > newest_date:
            newest_date = df["DatumLetzteAktualisierung"].max()

    # Save metadata along with data
    metadata_file = os.path.join(data_path, filenames["metadata"])
    metadata = create_datapackage_meta_json(
        newest_date,
        TECHNOLOGIES,
        data=["raw", "cleaned", "postprocessed"],
        json_serialize=False,
    )

    with open(metadata_file, "w", encoding="utf-8") as f:
        json.dump(metadata, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":

    pass



# Possible values for parameter 'data' with bulk download method
BULK_DATA = [
    "wind",
    "solar",
    "biomass",
    "hydro",
    "gsgk",
    "combustion",
    "nuclear",
    "gas",
    "storage",
    "electricity_consumer",
    "location",
    "market",
    "grid",
    "balancing_area",
    "permit",
    "deleted_units",
    "deleted_market_actors",
    "retrofit_units",
    "changed_dso_assignment",
    "storage_units",
]

# Possible values for parameter 'data' with API download method
API_DATA = [
    "wind",
    "solar",
    "biomass",
    "hydro",
    "gsgk",
    "combustion",
    "nuclear",
    "storage",
    "location",
    "permit",
]

# Technology related values of parameter 'data'
# Methods like Mastr.to_csv() must separate these from additional tables
TECHNOLOGIES = [
    "wind",
    "solar",
    "biomass",
    "hydro",
    "gsgk",
    "combustion",
    "nuclear",
    "storage",
]

# Data which are not related to a specific technology
ADDITIONAL_TABLES = [
    "balancing_area",
    "electricity_consumer",
    "gas_consumer",
    "gas_producer",
    "gas_storage",
    "gas_storage_extended",
    "grid_connections",
    "grids",
    "locations_extended",
    "market_actors",
    "market_roles",
    "permit",
    "deleted_units",
    "deleted_market_actors",
    "retrofit_units",
    "changed_dso_assignment",
    "storage_units",
]

# Possible data types for API download
API_DATA_TYPES = ["unit_data", "eeg_data", "kwk_data", "permit_data"]

# Possible location types for API download
API_LOCATION_TYPES = [
    "location_elec_generation",
    "location_elec_consumption",
    "location_gas_generation",
    "location_gas_consumption",
]

# Map bulk data to bulk download tables (XML file names)
BULK_INCLUDE_TABLES_MAP = {
    "wind": ["anlageneegwind", "einheitenwind"],
    "solar": ["anlageneegsolar", "einheitensolar"],
    "biomass": ["anlageneegbiomasse", "einheitenbiomasse"],
    "hydro": ["anlageneegwasser", "einheitenwasser"],
    "gsgk": [
        "anlageneeggeothermiegrubengasdruckentspannung",
        "einheitengeothermiegrubengasdruckentspannung",
    ],
    "combustion": ["anlagenkwk", "einheitenverbrennung"],
    "nuclear": ["einheitenkernkraft"],
    "storage": ["anlageneegspeicher", "einheitenstromspeicher"],
    "storage_units": ["anlagenstromspeicher"],
    "gas": [
        "anlagengasspeicher",
        "einheitengaserzeuger",
        "einheitengasspeicher",
        "einheitengasverbraucher",
    ],
    "electricity_consumer": ["einheitenstromverbraucher"],
    "location": ["lokationen"],
    "market": ["marktakteure", "marktrollen"],
    "grid": ["netzanschlusspunkte", "netze"],
    "balancing_area": ["bilanzierungsgebiete"],
    "permit": ["einheitengenehmigung"],
    "deleted_units": ["geloeschteunddeaktivierteeinheiten"],
    "deleted_market_actors": ["geloeschteunddeaktiviertemarktakteure"],
    "retrofit_units": ["ertuechtigungen"],
    "changed_dso_assignment": ["einheitenaenderungnetzbetreiberzuordnungen"],
}

# Map bulk data to database table names, for csv export
BULK_ADDITIONAL_TABLES_CSV_EXPORT_MAP = {
    "gas": [
        "gas_consumer",
        "gas_producer",
        "gas_storage",
        "gas_storage_extended",
    ],
    "electricity_consumer": ["electricity_consumer"],
    "location": ["locations_extended"],
    "market": ["market_actors", "market_roles"],
    "grid": ["grid_connections", "grids"],
    "balancing_area": ["balancing_area"],
    "permit": ["permit"],
    "deleted_units": ["deleted_units"],
    "deleted_market_actors": ["deleted_market_actors"],
    "retrofit_units": ["retrofit_units"],
    "changed_dso_assignment": ["changed_dso_assignment"],
}

# used to map the parameter options in open-mastr to the exact table class names in orm.py
ORM_MAP = {
    "wind": {
        "unit_data": "WindExtended",
        "eeg_data": "WindEeg",
        "permit_data": "Permit",
    },
    "solar": {
        "unit_data": "SolarExtended",
        "eeg_data": "SolarEeg",
        "permit_data": "Permit",
    },
    "biomass": {
        "unit_data": "BiomassExtended",
        "eeg_data": "BiomassEeg",
        "kwk_data": "Kwk",
        "permit_data": "Permit",
    },
    "combustion": {
        "unit_data": "CombustionExtended",
        "kwk_data": "Kwk",
        "permit_data": "Permit",
    },
    "gsgk": {
        "unit_data": "GsgkExtended",
        "eeg_data": "GsgkEeg",
        "kwk_data": "Kwk",
        "permit_data": "Permit",
    },
    "hydro": {
        "unit_data": "HydroExtended",
        "eeg_data": "HydroEeg",
        "permit_data": "Permit",
    },
    "nuclear": {"unit_data": "NuclearExtended", "permit_data": "Permit"},
    "storage": {
        "unit_data": "StorageExtended",
        "eeg_data": "StorageEeg",
        "permit_data": "Permit",
    },
    "gas_consumer": "GasConsumer",
    "gas_producer": "GasProducer",
    "gas_storage": "GasStorage",
    "gas_storage_extended": "GasStorageExtended",
    "electricity_consumer": "ElectricityConsumer",
    "locations_extended": "LocationExtended",
    "market_actors": "MarketActors",
    "market_roles": "MarketRoles",
    "grid_connections": "GridConnections",
    "grids": "Grids",
    "balancing_area": "BalancingArea",
    "permit": "Permit",
    "deleted_units": "DeletedUnits",
    "deleted_market_actors": "DeletedMarketActors",
    "retrofit_units": "RetrofitUnits",
    "changed_dso_assignment": "ChangedDSOAssignment",
    "storage_units": "StorageUnits",
}

UNIT_TYPE_MAP = {
    "Windeinheit": "wind",
    "Solareinheit": "solar",
    "Biomasse": "biomass",
    "Wasser": "hydro",
    "Geothermie": "gsgk",
    "Verbrennung": "combustion",
    "Kernenergie": "nuclear",
    "Stromspeichereinheit": "storage",
    "Gasspeichereinheit": "gas_storage",
    "Gasverbrauchseinheit": "gas_consumer",
    "Stromverbrauchseinheit": "electricity_consumer",
    "Gaserzeugungseinheit": "gas_producer",
    "Stromerzeugungslokation": "location_elec_generation",
    "Stromverbrauchslokation": "location_elec_consumption",
    "Gaserzeugungslokation": "location_gas_generation",
    "Gasverbrauchslokation": "location_gas_consumption",
}

TRANSLATIONS = {
    "RegisternummerAusland": "foreignRegisterNumber",
    "PumpbetriebKontinuierlichRegelbar": "continuousControlOfPumpOperation",
    "AuflagenAbschaltungSonstige": "requirementShutdownOther",
    "OrtAnZustelladresse": "cityForDeliveryAddress",
    "RegelzoneNetzanschlusspunkt": "controlZoneNetworkConnectionPoint",
    "EegMastrNummer": "eegMastrNumber",
    "AusschreibungZuschlag": "tenderAward",
    "NameGasspeicher": "gasStorageName",
    "BelieferungVonLetztverbrauchernGas": "supplyToFinalConsumersGas",
    "Registrierungsdatum": "registrationDate",
    "Einheitart": "unitType",
    "BiogasGaserzeugungskapazitaet": "biogasGasProductionCapacity",
    "FernsteuerbarkeitDv": "directMarketerRemoteControl",
    "reason": "reason",
    "GaussKruegerHoch": "gaussKruegerHigh",
    "DatumBeginnVorlaeufigenOderEndgueltigenStilllegung": "startOfTemporaryOrPermanentShutdown",
    "BelieferungHaushaltskundenStrom": "supplyHouseholdCustomersElectricity",
    "GeschlossenesVerteilnetz": "closedDistributionNetwork",
    "DatumWiederaufnahmeBetrieb": "resumptionOfOperationDate",
    "HausnummerAnZustelladresse": "houseNumberForDeliveryAddress",
    "NettonennleistungDeutschland": "netCapacityGermany",
    "DatumBeendigungVorlaeufigenStilllegung": "temporaryShutdownEndDate",
    "KwkMastrNummer": "kwkMastrNumber",
    "NetzMastrNummer": "networkMastrNumber",
    "Nuts2": "nuts2",
    "Yeic": "yeic",
    "request_date": "request_date",
    "Fax_nv": "fax_na",
    "AnzahlStromverbrauchseinheitenGroesser50Mw": "numberPowerConsumptionUnitsGreaterThan50Mw",
    "Registergericht": "commercialRegister",
    "Frist": "deadline",
    "Breitengrad": "latitude",
    "Weic_Na": "weic_na",
    "PilotAnlage": "pilotPlant",
    "MarktakteurNachname": "marketParticipantLastName",
    "MieterstromMeldedatum": "TenantPowerReportingDate",
    "VerhaeltnisReferenzertragErtrag10Jahre_nv": "referenceYieldRatio10YearsYield_na",
    "NameKraftwerksblock": "powerPlantBlockName",
    "Marktrollen": "marketRoles",
    "HauptausrichtungNeigungswinkel": "mainOrientationTiltAngle",
    "Name": "name",
    "Nettoengpassleistung": "netShortCircuitPower",
    "NameWindpark": "windFarmName",
    "Batterietechnologie": "batteryTechnology",
    "VerhaeltnisReferenzertragErtrag10Jahre": "referenceYieldRatio10YearsYield",
    "MieterstromZugeordnet": "TenantPowerAssigned",
    "ClusterOstsee": "clusterBalticSea",
    "BiogasDatumInanspruchnahmeFlexiPraemie": "biogasFlexiPremiumUtilizationDate",
    "UtmEast": "utmEast",
    "Adresszusatz": "addressAdditionalInfo",
    "Bundesland": "state",
    "EinheitSystemstatus": "unitSystemStatus",
    "Datum": "date",
    "WasserrechtAblaufdatum": "waterRightsExpirationDate",
    "ArtDesZuflusses": "typeOfInflow",
    "Technologie": "technology",
    "BiomethanErstmaligerEinsatz_nv": "biomethaneInitialUse_na",
    "Hausnummer": "houseNumber",
    "PostleitzahlAnZustelladresse": "postalCodeForDeliveryAddress",
    "AnlagenkennzifferAnlagenregister": "plantIdentificationNumberRegister",
    "WiederinbetriebnahmeDatum": "recommissioningDate",
    "MinderungStromerzeugung": "reductionElectricityGeneration",
    "Einsatzverantwortlicher": "usageResponsibleEntity",
    "AuflagenAbschaltungSchattenwurf": "requirementShutdownShadowThrow",
    "MaximaleEinspeicherleistung": "maximumInjectionCapacity",
    "Nutzungsbereich": "usageArea",
    "Umsatzsteueridentifikationsnummer_nv": "vatIdentificationNumber_na",
    "Lage": "position",
    "DatenQuelle": "dataSource",
    "AnlagenschluesselEeg": "plantKeyEeg",
    "Laengengrad": "longitude",
    "MarktakteurVorname": "marketParticipantFirstName",
    "HauptwirtdschaftszweigGruppe": "mainEconomicSectorGroup",
    "BiogasLeistungserhoehung": "biogasCapacityIncrease",
    "AnteiligNutzungsberechtigte": "proportionallyAuthorizedUsers",
    "Region": "region",
    "KontaktdatenMarktrolle": "marketRoleContactData",
    "DurchschnittlicherBrennwert": "averageCalorificValue",
    "data_type": "data_type",
    "NebenausrichtungNeigungswinkel": "secondaryOrientationTiltAngle",
    "VerknuepfteEinheit": "linkedUnit",
    "Lokationtyp": "locationType",
    "BestandteilGrenzkraftwerk": "partOfBorderPowerPlant",
    "NameDerTechnischenLokation": "technicalLocationName",
    "GemeinsamerWechselrichterMitSpeicher": "sharedInverterWithStorage",
    "UtmZonenwert": "utmZoneValue",
    "SpeMastrNummer": "speMastrNumber",
    "DatumDownload": "downloadDate",
    "AnlagenbetreiberMastrNummer": "plantOperatorMastrNumber",
    "BilanzierungsgebietNetzanschlusspunkt": "accountingAreaNetworkConnectionPoint",
    "Fax": "fax",
    "Netzbetreiberzuordnungen": "gridOperatorAssignments",
    "NetzreserveAbDatum": "gridReserveFromDate",
    "PrototypAnlage": "prototypePlant",
    "BiogasUmfangLeistungserhoehung": "biogasCapacityIncreaseScope",
    "Kraftwerksnummer_nv": "powerPlantNumber_na",
    "WasserrechtsNummer": "waterRightsNumber",
    "ErtuechtigungIds": "rehabilitationIds",
    "FernsteuerbarkeitDr": "thirdPartyRemoteControl",
    "Einspeisungsart": "feedInType",
    "Umsatzsteueridentifikationsnummer": "vatIdentificationNumber",
    "NetzanschlusspunktMastrNummer": "networkConnectionPointMastrNumber",
    "VerknuepfteErzeugungseinheiten": "linkedGenerationUnits",
    "Marktpartneridentifikationsnummer_nv": "marketPartnerIdentificationNumber_na",
    "Wassertiefe": "waterDepth",
    "WeicDisplayName": "weicDisplayName",
    "eegAnlagenschluessel": "eegPlantKey",
    "Nabenhoehe": "hubHeight",
    "VerhaeltnisErtragsschaetzungReferenzertrag_nv": "yieldEstimationReferenceYieldRatio_na",
    "AuflagenAbschaltungTierschutz": "requirementShutdownAnimalProtection",
    "AnlagenkennzifferAnlagenregister_nv": "plantIdentificationNumberRegister_nv",
    "BiogasDatumLeistungserhoehung": "biogasCapacityIncreaseDate",
    "InAnspruchGenommeneLandwirtschaftlichGenutzteFlaeche": "areaOfAgriculturalLandInUse",
    "Aktenzeichen": "fileReference",
    "NetzbetreiberpruefungStatus": "gridOperatorCheckStatus",
    "AnlageBetriebsstatus": "plantOperatingStatus",
    "Gemarkung": "plot",
    "Art": "type",
    "AnschlussAnHoechstOderHochSpannung": "connectionToHighVoltageOrExtraHighVoltage",
    "MaximaleGasbezugsleistung": "maximumGasPurchaseCapacity",
    "InAnspruchGenommeneFlaeche": "areaInUse",
    "BelieferungHaushaltskundenGas": "supplyHouseholdCustomersGas",
    "AnzahlModule": "numberOfModules",
    "Standort": "location",
    "NetzbetreiberMastrNummer": "gridOperatorMastrNumber",
    "ZugeordneteWirkleistungWechselrichter": "assignedReactivePowerInverter",
    "ElektrischeKwkLeistung": "electricalCombinedHeatAndPowerCapacity",
    "Rechtsform": "legalForm",
    "ArtAbschaltbareLast": "typeSwitchableLoad",
    "AnteilBeinflussbareLast": "proportionControllableLoad",
    "NetzanschlusspunktBezeichnung": "networkConnectionPointDescription",
    "Webseite_nv": "website_na",
    "AuflagenAbschaltungSchallimmissionsschutzTagsueber": "requirementShutdownNoiseProtectionDay",
    "BestandsanlageMastrNummer": "inventoryPlantMastrNumber",
    "Gasqualitaet": "gasQuality",
    "Kmu": "smes",
    "download_date": "download_date",
    "NameStromerzeugungseinheit": "powerGenerationUnitName",
    "AnzahlNetzanschlusspunkte": "numberOfNetworkConnectionPoints",
    "BiogasGaserzeugungskapazitaet_nv": "biogasGasProductionCapacity_na",
    "AuflagenAbschaltungSchallimmissionsschutzNachts": "requirementShutdownNoiseProtectionNight",
    "LandAnZustelladresse": "countryForDeliveryAddress",
    "VerhaeltnisErtragsschaetzungReferenzertrag": "yieldEstimationReferenceYieldRatio",
    "VerknuepfteEinheiten": "linkedUnits",
    "InanspruchnahmeZahlungNachEeg": "UseOfPaymentAccordingToEeg",
    "Marktgebiet": "marketArea",
    "Stromgrosshaendler": "electricityWholesaler",
    "AuflagenAbschaltungEiswurf": "requirementShutdownIceThrow",
    "LokationMastrNummer": "locationMastrNumber",
    "RegistrierungsnummerPvMeldeportal_nv": "registrationNumberPvReportingPortal_na",
    "Marktfunktion": "marketFunction",
    "Netzanschlusskapazitaet": "networkConnectionCapacity",
    "Bezeichnung": "designation",
    "FlurFlurstuecknummern": "plotParcelNumbers",
    "MastrNummernKombibetrieb": "mastrNumbersCombinedOperation",
    "SteigerungNettonennleistungKombibetrieb": "increaseInNetCapacityCombinedOperation",
    "ArtDerFlaeche": "typeOfArea",
    "Marktpartneridentifikationsnummer": "marketPartnerIdentificationNumber",
    "Nettonennleistung": "netCapacity",
    "Buergerenergie": "communityEnergy",
    "GaussKruegerRechts": "gaussKruegerRight",
    "ErtuechtigungIstZulassungspflichtig": "upgradeRequiresApproval",
    "Behoerde": "authority",
    "AnlageIstImKombibetrieb": "plantIsInCombinedOperation",
    "WasserrechtAblaufdatum_nv": "waterRightsExpirationDate_na",
    "Leistungsbegrenzung": "powerLimitation",
    "Gasgrosshaendler": "naturalGasWholesaler",
    "NameGasverbrauchsseinheit": "gasConsumptionUnitName",
    "Netzanschlusspunkte": "networkConnectionPoints",
    "Direktvermarktungsunternehmen": "directMarketingCompany",
    "Postleitzahl": "postalCode",
    "Speicherart": "storageType",
    "AcerCode_nv": "acerCode_na",
    "Notstromaggregat": "emergencyGenerator",
    "MaximaleEinspeiseleistung": "maximumFeedInPower",
    "EinheitlicheAusrichtungUndNeigungswinkel": "uniformOrientationAndTiltAngle",
    "Rotordurchmesser": "rotorDiameter",
    "Taetigkeitsende_nv": "activityEndDate_na",
    "Netz": "network",
    "GeplantesInbetriebnahmedatum": "plannedCommissioningDate",
    "Einheittyp": "unitCategory",
    "Land": "country",
    "Hoechstbemessungsleistung": "maximumRatedCapacity",
    "BiomethanErstmaligerEinsatz": "biomethaneInitialUse",
    "InstallierteLeistung": "installedCapacity",
    "AuflageAbschaltungLeistungsbegrenzung": "requirementShutdownPowerLimitation",
    "StrasseNichtGefunden": "streetNotFound",
    "BiogasInanspruchnahmeFlexiPraemie": "biogasFlexiPremiumUtilization",
    "NameStromverbrauchseinheit": "powerConsumptionUnitName",
    "KundenAngeschlossen": "customersConnected",
    "Typenbezeichnung": "modelDesignation",
    "ArtDerWasserkraftanlage": "typeOfHydroPowerPlant",
    "Schwarzstartfaehigkeit": "blackStartCapability",
    "Ort": "place",
    "Leistungserhoehung": "powerIncrease",
    "Energietraeger": "energySource",
    "Inbetriebnahmedatum": "commissioningDate",
    "FernsteuerbarkeitNb": "gridOperatorRemoteControl",
    "Messlokation": "measurementLocation",
    "VerhaeltnisReferenzertragErtrag5Jahre": "referenceYieldRatio5YearsYield",
    "HauptwirtdschaftszweigAbteilung": "mainEconomicSectorDivision",
    "MaximalNutzbaresArbeitsgasvolumen": "maximumUsableWorkingGasVolume",
    "Erzeugungsleistung": "generationCapacity",
    "Hersteller": "manufacturer",
    "AcDcKoppelung": "acDcCoupling",
    "NochInPlanung": "stillInPlanning",
    "DatumEndgueltigeStilllegung": "finalShutdownDate",
    "StrasseAnZustelladresse": "streetForDeliveryAddress",
    "AusschliesslicheVerwendungBiomasse": "exclusiveUseOfBiomass",
    "Frist_nv": "deadline_na",
    "DatumDesBetreiberwechsels": "operatorChangeDate",
    "WeitereBrennstoffe": "otherFuels",
    "EinheitBetriebsstatus": "unitOperationalStatus",
    "Registernummer": "registerNumber",
    "NameKraftwerk": "powerPlantName",
    "Meldedatum": "notificationDate",
    "Einsatzort": "placeOfUse",
    "MarktakteurMastrNummer": "marketParticipantMastrNumber",
    "NameGaserzeugungseinheit": "gasGenerationUnitName",
    "DatumBeginnVoruebergehendeStilllegung": "temporaryShutdownStartDate",
    "Strasse": "street",
    "Ertuechtigungsart": "upgradeType",
    "Gemeinde": "municipality",
    "SicherheitsbereitschaftAbDatum": "securityReadinessFromDate",
    "Email": "email",
    "AcerCode": "acerCode",
    "VerhaeltnisReferenzertragErtrag15Jahre": "referenceYieldRatio15YearsYield",
    "Hauptausrichtung": "mainOrientation",
    "DatumLetzeAktualisierung": "lastUpdateDate",
    "RegistrierungsnummerPvMeldeportal": "RegistrationNumberPvNotificationPortal",
    "AltAnlagenbetreiberMastrNummer": "oldPlantOperatorMastrNumber",
    "DatumBaubeginn": "constructionStartDate",
    "Hauptbrennstoff": "mainFuel",
    "MieterstromErsteZuordnungZuschlag": "TenantPowerFirstAllocationSurcharge",
    "additional_data_id": "additional_data_id",
    "AnzeigeEinerStilllegung": "shutdownNotification",
    "PraequalifiziertGemaessAblav": "prequalifiedAccordingToAblav",
    "EinheitDientDerStromerzeugung": "unitServesElectricityGeneration",
    "Bruttoleistung": "grossCapacity",
    "Marktrolle": "marketRole",
    "ZugeordnenteWirkleistungWechselrichter": "allocatedReactivePowerOfInverter",
    "MaximaleAusspeiseleistung": "maximumWithdrawalPower",
    "Taetigkeitsende": "activityEndDate",
    "Gemeindeschluessel": "municipalityKey",
    "eegZuschlagsnummer": "eegAuctionBidNumber",
    "Biomasseart": "biomassType",
    "AusschliesslicheVerwendungImKombibetrieb": "exclusiveUseInCombinedOperation",
    "Inselbetriebsfaehigkeit": "islandOperationCapability",
    "SonstigeRechtsform": "otherLegalForm",
    "DatumRegistrierungDesBetreiberwechsels": "operatorChangeRegistrationDate",
    "MaximaleAusspeicherleistung": "maximumWithdrawalCapacity",
    "Zuschlagnummer": "grantNumber",
    "UtmNorth": "utmNorth",
    "Weic_nv": "weic_na",
    "BundesnetzagenturBetriebsnummer_nv": "federalNetworkAgencyOperationNumber_na",
    "ClusterNordsee": "clusterNorthSea",
    "Firmenname": "companyName",
    "GenMastrNummer": "genMastrNumber",
    "Seelage": "locationAtSea",
    "Nachtkennzeichen": "nightIdentifier",
    "HausnummerAnZustelladresse_nv": "houseNumberForDeliveryAddress_na",
    "PumpbetriebLeistungsaufnahme": "pumpOperationPowerConsumption",
    "NetzbetreiberpruefungDatum": "gridOperatorCheckDate",
    "AdresszusatzAnZustelladresse": "addressAdditionalInfoForDeliveryAddress",
    "Zuschlagsnummer": "awardNumber",
    "technology": "technology",
    "ThermischeNutzleistung": "thermalUtilizationCapacity",
    "NutzbareSpeicherkapazitaet": "usableStorageCapacity",
    "ArtDerStilllegung": "shutdownType",
    "EegAnlagentyp": "eegPlantType",
    "Sparte": "sector",
    "Personenart": "personType",
    "HauptwirtdschaftszweigAbschnitt": "mainEconomicSectorSection",
    "RegistergerichtAusland": "foreignCommercialRegister",
    "Spannungsebene": "voltageLevel",
    "BilanzierungsgebietNetzanschlusspunktId": "balancingAreaNetworkConnectionPointId",
    "BundesnetzagenturBetriebsnummer": "federalNetworkAgencyOperationNumber",
    "NichtVorhandenInMigriertenEinheiten": "notFoundInMigratedUnits",
    "Ertuechtigung": "rehabilitation",
    "Speichername": "storageName",
    "EinheitMastrNummer": "unitMastrNumber",
    "PraequalifiziertFuerRegelenergie": "prequalifiedForControlEnergy",
    "VerhaeltnisReferenzertragErtrag15Jahre_nv": "referenceYieldRatio15YearsYield_na",
    "ZugeordneteGebotsmenge": "AssignedBidQuantity",
    "Pumpspeichertechnologie": "pumpedStorageTechnology",
    "Id": "id",
    "Webseite": "website",
    "DatumLetzteAktualisierung": "lastUpdateDate",
    "Taetigkeitsbeginn": "activityStartDate",
    "Nebenausrichtung": "secondaryOrientation",
    "location_type": "location_type",
    "HerstellerId": "manufacturerId",
    "Hausnummer_nv": "houseNumber_na",
    "ArtDerFlaecheIds": "typeOfAreaIds",
    "Rotorblattenteisungssystem": "bladeIcingSystem",
    "VerhaeltnisReferenzertragErtrag5Jahre_nv": "referenceYieldRatio5YearsYield_na",
    "BelieferungVonLetztverbrauchernStrom": "supplyToFinalConsumersElectricity",
    "EegInbetriebnahmedatum": "eegCommissioningDate",
    "Telefon": "phone",
    "Landkreis": "district",
    "LetzteAenderung": "lastChange",
    "Weic": "weic",
    "SpeicherMastrNummer": "storageMastrNumber",
    "Kraftwerksnummer": "powerPlantNumber",
    "WeitererHauptbrennstoff": "additionalMainFuel",
    "id": "id",
    "HausnummerNichtGefunden": "houseNumberNotFound",
    "Anlagenbetreiber": "plantOperator",
    "RegistrierungsdatumMarktakteur": "marketParticipantRegistrationDate",
    "MieterstromRegistrierungsdatum": "TenantPowerRegistrationDate",
    "MastrNummer": "mastrNumber",
    "Kuestenentfernung": "distanceToCoast",
    "eegAusschreibungZuschlag": "eegAuctionBidAward",
    "DatumUeberfuehrungInReserve": "dateTransferToReserve",
    "ReserveartNachDemEnWG": "typeOfReserveFromEnWG",
    "WebportalDesNetzbetreibers": "webPortalGridOperator",
    "RegisternummerPraefix": "registerNumberPrefix",
}



import os
import sqlite3
import sys
from datetime import datetime
from os.path import expanduser
from zipfile import ZipFile

import numpy as np
import pandas as pd
import pytest
from sqlalchemy import create_engine, inspect
from sqlalchemy.sql import text

from open_mastr.utils import orm
from open_mastr.utils.orm import RetrofitUnits, NuclearExtended, tablename_mapping
from open_mastr.xml_download.utils_write_to_database import (
    add_missing_columns_to_table,
    add_zero_as_first_character_for_too_short_string,
    cast_date_columns_to_string,
    change_column_names_to_orm_format,
    correct_ordering_of_filelist,
    create_database_table,
    extract_sql_table_name,
    extract_xml_table_name,
    is_date_column,
    is_first_file,
    is_table_relevant,
    process_table_before_insertion,
    read_xml_file,
    add_table_to_non_sqlite_database,
    add_table_to_sqlite_database,
    interleave_files,
)

# Check if xml file exists
_xml_file_exists = False
_xml_folder_path = os.path.join(expanduser("~"), ".open-MaStR", "data", "xml_download")
if os.path.isdir(_xml_folder_path):
    for entry in os.scandir(path=_xml_folder_path):
        if "Gesamtdatenexport" in entry.name:
            _xml_file_exists = True


# Silence ValueError caused by logger https://github.com/pytest-dev/pytest/issues/5502
@pytest.fixture(autouse=True)
def capture_wrap():
    sys.stderr.close = lambda *args: None
    sys.stdout.close = lambda *args: None
    yield


@pytest.fixture(scope="module")
def zipped_xml_file_path():
    zipped_xml_file_path = None
    for entry in os.scandir(path=_xml_folder_path):
        if "Gesamtdatenexport" in entry.name:
            zipped_xml_file_path = os.path.join(_xml_folder_path, entry.name)

    return zipped_xml_file_path


@pytest.fixture(scope="module")
def con_testdb():
    testdb_file_path = os.path.join(
        expanduser("~"), ".open-MaStR", "data", "sqlite", "test-open-mastr.db"
    )
    # Create testdb
    con_testdb = sqlite3.connect(testdb_file_path)
    yield con_testdb
    con_testdb.close()
    # Remove testdb
    os.remove(testdb_file_path)


@pytest.fixture(scope="module")
def engine_testdb():
    testdb_file_path = os.path.join(
        expanduser("~"), ".open-MaStR", "data", "sqlite", "test-open-mastr.db"
    )
    testdb_url = f"sqlite:///{testdb_file_path}"
    yield create_engine(testdb_url)


def test_extract_xml_table_name():
    file_name = "Netzanschlusspunkte_31.xml"
    assert extract_xml_table_name(file_name) == "netzanschlusspunkte"


def text_extract_sql_table_name():
    xml_table_name = "netzanschlusspunkte"
    assert extract_sql_table_name(xml_table_name) == "network_connection_points"


def test_is_table_relevant():
    include_tables = ["anlagengasspeicher", "marktakteure"]
    assert is_table_relevant("anlagengasspeicher", include_tables) is True
    assert is_table_relevant("netzanschlusspunkte", include_tables) is False


def test_create_database_table(engine_testdb):
    orm.Base.metadata.create_all(engine_testdb)
    xml_table_name = "einheitenkernkraft"
    sql_table_name = "nuclear_extended"

    create_database_table(engine_testdb, xml_table_name)

    assert inspect(engine_testdb).has_table(sql_table_name) is True


def test_is_first_file():
    assert is_first_file("EinheitenKernkraft.xml") is True
    assert is_first_file("EinheitenKernkraft_1.xml") is True
    assert is_first_file("EinheitenKernkraft_2.xml") is False


def test_cast_date_columns_to_string():
    initial_df = pd.DataFrame(
        {
            "EegMastrNummer": [1, 2, 3],
            "Registrierungsdatum": [
                datetime(2024, 3, 11).date(),
                datetime(1999, 2, 1).date(),
                np.datetime64("nat"),
            ],
            "DatumLetzteAktualisierung": [
                datetime(2022, 3, 22),
                datetime(2020, 1, 2, 10, 12, 46),
                np.datetime64("nat"),
            ],
        }
    )
    expected_df = pd.DataFrame(
        {
            "EegMastrNummer": [1, 2, 3],
            "Registrierungsdatum": ["2024-03-11", "1999-02-01", np.nan],
            "DatumLetzteAktualisierung": [
                "2022-03-22 00:00:00.000000",
                "2020-01-02 10:12:46.000000",
                np.nan,
            ],
        }
    )

    pd.testing.assert_frame_equal(
        expected_df, cast_date_columns_to_string("anlageneegwasser", initial_df)
    )


def test_is_date_column():
    columns = RetrofitUnits.__table__.columns.items()
    df = pd.DataFrame(
        {
            "Id": [1],
            "DatumLetzteAktualisierung": [datetime(2022, 3, 22)],
            "WiederinbetriebnahmeDatum": [datetime(2024, 3, 11).date()],
        }
    )

    date_column = list(filter(lambda col: col[0] == "Id", columns))[0]
    assert is_date_column(date_column, df) is False

    datetime_column = list(
        filter(lambda col: col[0] == "DatumLetzteAktualisierung", columns)
    )[0]
    assert is_date_column(datetime_column, df) is True

    date_column = list(
        filter(lambda col: col[0] == "WiederinbetriebnahmeDatum", columns)
    )[0]
    assert is_date_column(date_column, df) is True


def test_correct_ordering_of_filelist():
    filelist = [
        "Solar_1.xml",
        "Solar_10.xml",
        "Solar_11.xml",
        "Solar_2.xml",
        "Solar_3.xml",
        "Solar_4.xml",
        "Solar_5.xml",
        "Solar_6.xml",
        "Solar_7.xml",
        "Solar_8.xml",
        "Solar_9.xml",
        "Wind_1.xml",
        "Wind_2.xml",
        "Wind_3.xml",
    ]
    filelist_2 = ["Solar_01.xml", "Solar_02.xml", "Solar_10.xml", "Wind_01.xml"]

    filelist_corrected = correct_ordering_of_filelist(filelist)
    filelist_2_corrected = correct_ordering_of_filelist(filelist_2)

    assert filelist_corrected == [
        "Solar_1.xml",
        "Solar_2.xml",
        "Solar_3.xml",
        "Solar_4.xml",
        "Solar_5.xml",
        "Solar_6.xml",
        "Solar_7.xml",
        "Solar_8.xml",
        "Solar_9.xml",
        "Solar_10.xml",
        "Solar_11.xml",
        "Wind_1.xml",
        "Wind_2.xml",
        "Wind_3.xml",
    ]
    assert filelist_2_corrected == [
        "Solar_01.xml",
        "Solar_02.xml",
        "Solar_10.xml",
        "Wind_01.xml",
    ]


@pytest.mark.skipif(
    not _xml_file_exists, reason="The zipped xml file could not be found."
)
def test_read_xml_file(zipped_xml_file_path):
    with ZipFile(zipped_xml_file_path, "r") as f:
        df = read_xml_file(f, "EinheitenKernkraft.xml")

    assert df.shape[0] > 0

    # Since the file is from the latest download, its content can vary over time. To make sure that the table is
    # correctly created, we check that all of its columns are associated are included in our mapping.
    for column in df.columns:
        if column in tablename_mapping["einheitenkernkraft"]["replace_column_names"]:
            column = tablename_mapping["einheitenkernkraft"]["replace_column_names"][
                column
            ]
        assert column in NuclearExtended.__table__.columns.keys()


def test_add_zero_as_first_character_for_too_short_string():
    # Prepare
    df_raw = pd.DataFrame(
        {"ID": [0, 1, 2], "Gemeindeschluessel": [9162000, np.nan, 19123456]}
    )
    df_correct = pd.DataFrame(
        {"ID": [0, 1, 2], "Gemeindeschluessel": ["09162000", np.nan, "19123456"]}
    )

    # Act
    df_edited = add_zero_as_first_character_for_too_short_string(df_raw)
    # Assert
    pd.testing.assert_frame_equal(df_edited, df_correct)


def test_change_column_names_to_orm_format():
    initial_df = pd.DataFrame(
        {
            "VerknuepfteEinheitenMaStRNummern": ["test1", "test2"],
            "NetzanschlusspunkteMaStRNummern": [1, 2],
        }
    )
    expected_df = pd.DataFrame(
        {
            "VerknuepfteEinheiten": ["test1", "test2"],
            "Netzanschlusspunkte": [1, 2],
        }
    )

    pd.testing.assert_frame_equal(
        expected_df, change_column_names_to_orm_format(initial_df, "lokationen")
    )


@pytest.mark.skipif(
    not _xml_file_exists, reason="The zipped xml file could not be found."
)
def test_process_table_before_insertion(zipped_xml_file_path):
    bulk_download_date = datetime.now().date().strftime("%Y%m%d")
    initial_df = pd.DataFrame(
        {
            "Gemeindeschluessel": [9162000, 19123456],
            "Postleitzahl": [1234, 54321],
            "NameKraftwerk": ["test1", "test2"],
            "LokationMaStRNummer": ["test3", "test4"],
        }
    )
    expected_df = pd.DataFrame(
        {
            "Gemeindeschluessel": ["09162000", "19123456"],
            "Postleitzahl": ["01234", "54321"],
            "NameKraftwerk": ["test1", "test2"],
            "LokationMastrNummer": ["test3", "test4"],
            "DatenQuelle": ["bulk", "bulk"],
            "DatumDownload": [bulk_download_date, bulk_download_date],
        }
    )

    pd.testing.assert_frame_equal(
        expected_df,
        process_table_before_insertion(
            initial_df,
            "einheitenkernkraft",
            zipped_xml_file_path,
            bulk_download_date,
            bulk_cleansing=False,
        ),
    )


def test_add_missing_columns_to_table(engine_testdb):
    with engine_testdb.connect() as con:
        with con.begin():
            # We must recreate the table to be sure that the new colum is not present.
            con.execute(text("DROP TABLE IF EXISTS gas_consumer"))
            create_database_table(engine_testdb, "einheitengasverbraucher")

            initial_data_in_db = pd.DataFrame(
                {
                    "EinheitMastrNummer": ["id1"],
                    "DatumLetzteAktualisierung": [datetime(2022, 2, 2)],
                }
            )
            initial_data_in_db.to_sql(
                "gas_consumer", con=con, if_exists="append", index=False
            )

    add_missing_columns_to_table(
        engine_testdb, "einheitengasverbraucher", ["NewColumn"]
    )

    expected_df = pd.DataFrame(
        {
            "EinheitMastrNummer": ["id1"],
            "DatumLetzteAktualisierung": [datetime(2022, 2, 2)],
            "NewColumn": [None],
        }
    )
    with engine_testdb.connect() as con:
        with con.begin():
            actual_df = pd.read_sql_table("gas_consumer", con=con)
            # The actual_df will contain more columns than the expected_df, so we can't use assert_frame_equal.
            assert expected_df.index.isin(actual_df.index).all()


@pytest.mark.parametrize(
    "add_table_to_database_function",
    [add_table_to_sqlite_database, add_table_to_non_sqlite_database],
)
def test_add_table_to_sqlite_database(engine_testdb, add_table_to_database_function):
    with engine_testdb.connect() as con:
        with con.begin():
            # We must recreate the table to be sure that no other data is present.
            con.execute(text("DROP TABLE IF EXISTS gsgk_eeg"))
            create_database_table(
                engine_testdb, "anlageneeggeothermiegrubengasdruckentspannung"
            )

    df = pd.DataFrame(
        {
            "Registrierungsdatum": ["2022-02-02", "2024-03-20"],
            "EegMastrNummer": ["id1", "id2"],
            "DatumLetzteAktualisierung": [
                "2022-12-02 10:10:10.000300",
                "2024-10-10 00:00:00.000000",
            ],
            "AusschreibungZuschlag": [True, False],
            "Netzbetreiberzuordnungen": ["test1", "test2"],
            "InstallierteLeistung": [1.0, 100.4],
        }
    )
    expected_df = pd.DataFrame(
        {
            "InstallierteLeistung": [1.0, 100.4],
            "AnlageBetriebsstatus": [None, None],
            "Registrierungsdatum": [datetime(2022, 2, 2), datetime(2024, 3, 20)],
            "EegMastrNummer": ["id1", "id2"],
            "Meldedatum": [np.datetime64("NaT"), np.datetime64("NaT")],
            "DatumLetzteAktualisierung": [
                datetime(2022, 12, 2, 10, 10, 10, 300),
                datetime(2024, 10, 10),
            ],
            "EegInbetriebnahmedatum": [np.datetime64("NaT"), np.datetime64("NaT")],
            "VerknuepfteEinheit": [None, None],
            "AnlagenschluesselEeg": [None, None],
            "AusschreibungZuschlag": [True, False],
            "AnlagenkennzifferAnlagenregister": [None, None],
            "AnlagenkennzifferAnlagenregister_nv": [None, None],
            "Netzbetreiberzuordnungen": ["test1", "test2"],
            "DatenQuelle": [None, None],
            "DatumDownload": [np.datetime64("NaT"), np.datetime64("NaT")],
        }
    )

    add_table_to_database_function(
        df, "anlageneeggeothermiegrubengasdruckentspannung", "gsgk_eeg", engine_testdb
    )
    with engine_testdb.connect() as con:
        with con.begin():
            pd.testing.assert_frame_equal(
                expected_df, pd.read_sql_table("gsgk_eeg", con=con)
            )


def test_interleave_files():
    input_data = [
        ("AnlagenEegBiomasse.xml", "anlageneegbiomasse", "anlageneegbiomasse"),
        ("AnlagenEegSolar_1.xml", "anlageneegsolar", "anlageneegsolar"),
        ("AnlagenEegSolar_2.xml", "anlageneegsolar", "anlageneegsolar"),
        ("AnlagenEegWind_1.xml", "anlageneegwind", "anlageneegwind"),
        ("AnlagenEegWind_2.xml", "anlageneegwind", "anlageneegwind"),
        ("AnlagenEegWind_3.xml", "anlageneegwind", "anlageneegwind"),
        ("Bilanzierungsgebiete.xml", "bilanzierungsgebiete", "bilanzierungsgebiete"),
    ]
    assert interleave_files(input_data) == [
        ("AnlagenEegBiomasse.xml", "anlageneegbiomasse", "anlageneegbiomasse"),
        ("AnlagenEegSolar_1.xml", "anlageneegsolar", "anlageneegsolar"),
        ("AnlagenEegWind_1.xml", "anlageneegwind", "anlageneegwind"),
        ("Bilanzierungsgebiete.xml", "bilanzierungsgebiete", "bilanzierungsgebiete"),
        ("AnlagenEegSolar_2.xml", "anlageneegsolar", "anlageneegsolar"),
        ("AnlagenEegWind_2.xml", "anlageneegwind", "anlageneegwind"),
        ("AnlagenEegWind_3.xml", "anlageneegwind", "anlageneegwind"),
    ]
